;; Functions to handle the calculations for model predictions,  
;; geometry information, and some updated game interface code.

;; Parameters and constants
(defparameter *hex-radius* 175) ; in pixels
(defparameter *fortress-radius* 40) ; in pixels
(defparameter *acceleration* 0.05) ; pixels/gametick^2
(defparameter *radian* (/ 180 pi))

(defparameter *sgd-alpha* 0.1) ; stochastic gradient descent learning rate
;; (defparameter *alpha* 0.2)
;; (defparameter *factor* 1)

(defparameter *points* 0)
(defparameter *game-points* '())
(defparameter *moment-by-moment-data* '())
;(defparameter *experiment-points* '())
;(defparameter *params* '())

;; These need to be stored by model not globally and cleared at game start
(defparameter *shells* (make-hash-table)) ; '())
(defparameter *missiles* (make-hash-table)) ; '())
(defparameter *fortress-alive* (make-hash-table)) ; t)
(defparameter *self-alive* (make-hash-table)) ; t)

;; Functions
(defun vector-length (vector-coordinates)
  (sqrt (apply #'+ (mapcar (lambda (x) (* x x)) vector-coordinates))))

(defun sqr (x)
  (* x x))

(defun to-radians (angle)
  (/ angle *radian*))

(defun to-degs (angle)
  (* angle *radian*))

(defun angle (x1 y1 x2 y2)
  ;; angle of vector from first to second point?
  (let* ((run (- x2 x1))
         (rise (- y2 y1))
         (a (if (not (zerop run))  (* (atan (/ rise run)) *radian*))))
    (cond ((zerop run) (if (>= rise 0) 90 270))
          ((> run 0) (if (>= rise 0) a (+ 360 a)))
          (t (if (>= rise 0) (+ 180 a) (+ 180 a))))))

(defun angle-difference (alpha beta)
  (let ((diff (- alpha beta)))
    (cond ((and (<= 0 diff) (>= 360 diff)) diff)
          ((> 0 diff) (angle-difference (+ 360 alpha) beta))
          ((< 360 diff) (angle-difference (- alpha 360) beta)))))

(defun angle-difference2 (alpha beta)
  (let ((diff (angle-difference alpha beta)))
    (if (> diff 180)
	(- diff 360)
	diff)))

(defun angles-within (alpha beta interval)
  (or (<= (angle-difference alpha beta) interval) (<= (angle-difference beta alpha) interval)))

(defun angle-offset (ang1 ang2)
  ;;calculates the difference between 2 angles dealing with looping at 360
  (let* ((offset (mod (- ang1 ang2) 360)))
    (if (> offset 180) (- 360 offset) offset)))

(defun closest-angle (angle1 angle2)
  (let ((angle-diff (angle-difference angle1 angle2)))
    (if (> angle-diff 180) (- 360 angle-diff) angle-diff)))

(defun remove-every-other (lst)
  (if (< (length lst) 2)
      lst
      (cons (first lst) (remove-every-other (cddr lst)))))

(defun clockwisep (angle1 angle2)
  (>= (angle-difference angle1 angle2) 180))

(defun counterclockwisep (angle1 angle2)
  (< (angle-difference angle1 angle2) 180))

(defun acceleratingp (vx vy aim)
  (angles-within (angle 0 0 vx vy) aim 90))

(defun time-to-radius (radius x y vx vy time)
  (if (> time 500)
      time
      (let ((current-radius (vector-length (list x y))))
	(if (> current-radius radius)
	    time
	    (time-to-radius radius (+ x vx) (+ y vy) vx vy (+ time 1))))))

(defun default-chunks (lis) (eval (cons 'add-dm (mapcar #'(lambda (x) (list x 'isa 'chunk)) lis))))

(defun angle-toward-fortress (x y)
  (angle-difference (angle 0 0 x y) 180))

(defun avoid-fortress-angle (x y fortress-back-angle) ;; (x y vx vy fortress-back-angle)
  ;; target flight direction is towards back of fortress
  (let* ((ship-position-angle (angle 0 0 x y))
         (fortress-ship-relative (if (clockwisep fortress-back-angle ship-position-angle) 'clockwise 'counterclockwise))
         (target-flight-direction (if (equalp fortress-ship-relative 'clockwise) (angle-difference ship-position-angle 60) (angle-difference ship-position-angle -60)))
         )
    target-flight-direction))

(defun compute-aim (target-fd vx vy target-speed)
  ;; takes into account that the minimum change in speed
  ;; is ~ 0.3 pixels / tick (based on a tap thrust of 90 ms)
  (let* ((target-fd-rad (to-radians target-fd))
         (target-vx (* target-speed (cos target-fd-rad)))
         (target-vy (* target-speed (sin target-fd-rad)))
	 (flight-dir (angle 0 0 vx vy))
         (delta-vx (- target-vx vx))
         (delta-vy (- target-vy vy))
         (delta-speed (vector-length (list delta-vx delta-vy))))
    (if (angles-within target-fd flight-dir 10)
	flight-dir
	(if (> delta-speed 0.3)
            (angle-difference (to-degs (atan delta-vy delta-vx)) 360)
            (let* ((target-vx-new (* (+ target-speed 0.1) (cos target-fd-rad)))
		   (target-vy-new (* (+ target-speed 0.1) (sin target-fd-rad)))
		   (delta-vx-new (- target-vx-new vx))
		   (delta-vy-new (- target-vy-new vy)))
              (angle-difference (to-degs (atan delta-vy-new delta-vx-new)) 360))))))

(defun behind-angle (ship-x ship-y ship-vx ship-vy fortress-back-angle shooter-distance horizon)
  (let* ((pred-x (+ ship-x (* ship-vx horizon)))
	 (pred-y (+ ship-y (* ship-vy horizon)))
	 (distance-from-center (sqrt (+ (sqr pred-x) (sqr pred-y))))
	 (ship-angle (angle 0 0 pred-x pred-y))
         (current-minus-target (angle-difference ship-angle fortress-back-angle)))
    (cond
      ((< distance-from-center 60)
       (if (> current-minus-target 0)
           (angle-difference ship-angle -45)
           (angle-difference ship-angle 45)))
      ((<= current-minus-target 30)
       (angle ship-x ship-y
              (* shooter-distance (cos (to-radians (angle-difference ship-angle 90))))
              (* shooter-distance (sin (to-radians (angle-difference ship-angle 90))))))
      ((> current-minus-target 330)
       (angle ship-x ship-y
              (* shooter-distance (cos (to-radians (angle-difference ship-angle -90))))
              (* shooter-distance (sin (to-radians (angle-difference ship-angle -90))))))
      ((or (and (<= current-minus-target 90) (> current-minus-target 30) )
           (and (> current-minus-target 270) (<= current-minus-target 330)))
       (angle ship-x ship-y
              (* shooter-distance (cos (to-radians fortress-back-angle)))
              (* shooter-distance (sin (to-radians fortress-back-angle)))))
      ((and (> current-minus-target 90) (<= current-minus-target 180))
       (angle ship-x ship-y
              (* shooter-distance (cos (to-radians (angle-difference ship-angle 90))))
              (* shooter-distance (sin (to-radians (angle-difference ship-angle 90))))))
      ((and (> current-minus-target 180) (<= current-minus-target 270))
       (angle ship-x ship-y
              (* shooter-distance (cos (to-radians (angle-difference ship-angle -90))))
              (* shooter-distance (sin (to-radians (angle-difference ship-angle -90)))))))))

(defun outside-angle (x y vx vy horizon target-speed)
  (let* ((x-pred (+ x (* horizon vx)))
         (y-pred (+ y (* horizon vy)))
         (radius (vector-length (list x-pred y-pred)))
         (vx-outward (* (/ x-pred radius) target-speed))
         (vy-outward (* (/ y-pred radius) target-speed))
         (flight-direction (angle 0 0 vx vy))
         (ship-angle (angle 0 0 x-pred y-pred))
	 )
    (if (< (time-to-radius *hex-radius* x-pred y-pred vx vy 0)
           (time-to-radius *hex-radius* x-pred y-pred vx-outward vy-outward 0))
        flight-direction
        ship-angle)))

(defun bait-inside-angle (x y vx vy bait-distance)
  ;; choose direction perpendicular to radius in direction of current velocity
  (let* ((current-fd (angle 0 0 vx vy))
	 (current-loc-angle (angle 0 0 x y))
	 (target-location-angle (if (angles-within current-fd (angle-difference current-loc-angle 90) 90)
				    (angle-difference current-loc-angle 90)
				    (angle-difference current-loc-angle -90)))
	 (target-x (* bait-distance (cos (to-radians target-location-angle))))
	 (target-y (* bait-distance (sin (to-radians target-location-angle)))))
    (angle x y target-x target-y)))

(defun bait-angle (x y vx vy other-x other-y shell-x shell-y bait-distance)
  (let* ((ship-angle (angle 0 0 x y)) 
         (ship-distance (vector-length (list x y))) 
         (target-angle-counterclockwise (angle-difference ship-angle -90)) 
         (target-angle-clockwise (angle-difference ship-angle 90)) 
         (shell-angle (angle 0 0 shell-x shell-y))
         (shell-relative-position (if (clockwisep shell-angle ship-angle) 'clockwise 'counterclockwise))
         (speed (vector-length (list vx vy))) 
         (movement-relative-direction (if (clockwisep (angle 0 0 (+ x vx) (+ y vy)) ship-angle) 'clockwise 'counterclockwise)) 
         (other-ship-angle (angle 0 0 other-x other-y)) 
         (other-relative-direction (if (clockwisep other-ship-angle ship-angle) 'clockwise 'counterclockwise)) ;counterclockwise, but just a matter of less than 1 degree
         (final-target-angle
           (if (not (equalp shell-x 1000))
               (if (equalp shell-relative-position 'clockwise)
                   target-angle-counterclockwise
                   target-angle-clockwise) ;; if missiles present, fly in direction away from missiles
               (if (> ship-distance *hex-radius*)
                   (if (< (angle-difference ship-angle other-ship-angle) 100)
		       (if (equalp other-relative-direction 'clockwise)
			   target-angle-counterclockwise
			   target-angle-clockwise)
		       (if (equalp movement-relative-direction 'clockwise)
			   target-angle-clockwise
			   target-angle-counterclockwise))
                   (if (> speed 0.1)
                       (if (equalp movement-relative-direction 'clockwise)
                           target-angle-clockwise
                           target-angle-counterclockwise) ;; if no missiles are moving, change in direction closer to flight direction
                       (if (equal (act-r-random 2) 1)
                           target-angle-clockwise
                           target-angle-counterclockwise)) ;; if not moving, pick randomly one of the two possible targets
                   ))))
    (angle x y
           (* bait-distance (cos (to-radians final-target-angle)))
           (* bait-distance (sin (to-radians final-target-angle))))))

(defun stay-inside-angle (x y vx vy bait-distance)
  ;; choose direction perpendicular to radius in direction of current velocity
  (let* ((current-fd (angle 0 0 vx vy))
         (current-loc-angle (angle 0 0 x y))
         (target-location-angle (if (angles-within current-fd (angle-difference current-loc-angle 90) 90)
                                    (angle-difference current-loc-angle 90)
                                    (angle-difference current-loc-angle -90)))
         (target-x (* bait-distance (cos (to-radians target-location-angle))))
         (target-y (* bait-distance (sin (to-radians target-location-angle)))))
    (angle x y target-x target-y)))

(defun values-equal (dimension current target)
  (case dimension
    (location (or (and (equalp target 'insideHex) (equalp current 'behindFortress))
		  (equalp current target)))
    (other-location (or (and (equalp target 'insideHex) (equalp current 'behindFortress))
			(and (equalp target 'insideHex) (equalp current 'atBoundary))
			(equalp current target)))
    (fortress-target (equalp current target))
    (fortress-back (equalp current target))))

(defun location (x y fortress-angle fortress-alive)
  (labels ((behind-fortress-p (fortress-angle ship-x ship-y)
	     (let ((distance-from-center (sqrt (+ (sqr ship-x) (sqr ship-y)))))
	       (and
		(angles-within fortress-angle (angle-toward-fortress ship-x ship-y) 35)
		(<= distance-from-center (- *hex-radius* 10))
		(>= distance-from-center *fortress-radius*))))

	   (at-fortress-p (ship-x ship-y fortress-alive)
	     (let ((radius (sqrt (+ (sqr ship-x) (sqr ship-y)))))
	       (and fortress-alive (< radius *fortress-radius*))))

	   (at-outer-p (ship-x ship-y)
	     (or (<= ship-x -349) (>= ship-x 349) (<= ship-y -309) (>= ship-y 309))))

    (let ((distance-from-center (sqrt (+ (sqr x) (sqr y)))))
      (cond
	((at-outer-p x y) 'crashedIntoOuter)
	((> distance-from-center (+ *hex-radius* 17)) 'outsideHex)
	((> distance-from-center (- *hex-radius* 10)) 'atBoundary)
	((at-fortress-p x y fortress-alive) 'crashedIntoFortress)
	((behind-fortress-p fortress-angle x y) 'behindFortress)
	((<= distance-from-center *hex-radius*) 'insideHex)))))

(defun predict-back-orientation (other-x other-y other-vx other-vy horizon)
  (let* ((anticip-other-x (+ other-x (* horizon other-vx)))
         (anticip-other-y (+ other-y (* horizon other-vy)))
         (other-angle (angle 0 0 anticip-other-x anticip-other-y)))
    (angle-difference other-angle 180)))

(defun predicted-location (x y vx vy ox oy ovx ovy fortress-alive horizon)
  (labels ((at-outer-p (ship-x ship-y)
	     (or (<= ship-x -349) (>= ship-x 349) (<= ship-y -309) (>= ship-y 309)))
	   (future-at-outer-p (ship-x ship-y ship-vx ship-vy)
	     (at-outer-p (+ ship-x (* ship-vx horizon))
			 (+ ship-y (* ship-vy horizon))))

	   (behind-fortress-p (fortress-angle ship-x ship-y)
	     (let ((distance-from-center (sqrt (+ (sqr ship-x) (sqr ship-y)))))
	       (and
		(angles-within fortress-angle (angle-toward-fortress ship-x ship-y) 35)
		(<= distance-from-center (- *hex-radius* 10))
		(>= distance-from-center *fortress-radius*))))

	   (future-behind-fortress-p (ship-x ship-y ship-vx ship-vy other-x other-y other-vx other-vy horizon)
	     (behind-fortress-p (angle-difference
				 (predict-back-orientation other-x other-y other-vx other-vy horizon) 180)
				(+ ship-x (* horizon ship-vx))
				(+ ship-y (* horizon ship-vy))))

	   (at-fortress-p (ship-x ship-y fortress-alive)
	     (let ((radius (sqrt (+ (sqr ship-x) (sqr ship-y)))))
	       (and fortress-alive (< radius *fortress-radius*))))

	   (future-at-fortress-p (ship-x ship-y ship-vx ship-vy fortress-alive)
	     (at-fortress-p (+ ship-x (* ship-vx horizon))
			    (+ ship-y (* ship-vy horizon))
			    fortress-alive)))
    (let* ((future-x (+ x (* horizon vx)))
           (future-y (+ y (* horizon vy)))
           (future-distance-from-center (vector-length (list future-x future-y))))
      (cond
       ((and (> (sqrt (+ (sqr vx) (sqr vy))) .14) 
             (future-at-outer-p x y vx vy)) 'crashedIntoOuter)
       ((> future-distance-from-center (+ *hex-radius* 17)) 'outsideHex)
       ((> future-distance-from-center (- *hex-radius* 10)) 'atBoundary)
       ((future-at-fortress-p x y vx vy fortress-alive) 'crashedIntoFortress)
       ((and (equal fortress-alive 'yes) (future-behind-fortress-p x y vx vy ox oy ovx ovy horizon)) 'behindFortress)
       ((<= future-distance-from-center *hex-radius*) 'insideHex)))))

(defun predicted-location-union (x y vx vy ox oy ovx ovy fortress-alive horizon)
  (let ((current-location (predicted-location x y vx vy ox oy ovx ovy fortress-alive 0))
	(next-location (predicted-location x y vx vy ox oy ovx ovy fortress-alive 1)))
    (if (or (not (equal current-location next-location))
	    (<= horizon 1))
	next-location
	(predicted-location-union (+ x vx) (+ y vy) vx vy (+ ox ovx) (+ oy ovy) ovx ovy fortress-alive (- horizon 1)))))

(defun predict-aim-at-fortress (x y vx vy horizon)
  (angle 0 0 (- (+ x (* horizon vx))) (- (+ y (* horizon vy)))))

(defun predict-direction (vx vy aim horizon)
  (angle 0 0 (+ vx (* *acceleration* horizon (cos (/ aim *radian*))))
	 (+ vy (* *acceleration* horizon (sin (/ aim *radian*))))))

(defun predict-speed(vx vy aim horizon)
  (vector-length (list (+ vx (* *acceleration* horizon (cos (/ aim *radian*))))
		      (+ vy (* *acceleration* horizon (sin (/ aim *radian*)))))))

(defun direction-close (current target predicted)
  (or
   ;(angles-within current target 10)
   (and (> current target) (< predicted target))
   (and (< current target) (> predicted target))))

(defun speed-close (current target)
  (< (abs (- current target)) 0.14))

;; Added this to make it easy to change
(defvar *bait-direction-angle* 55) ;; originally was 55
(defvar *bait-direction-speed* .05) ;; original value .05

(defun flight-direction-bait-p (x y vx vy)
  (let ((target (angle-difference (angle 0 0 x y) 90))
	(speed (vector-length (list vx vy)))
	(current (angle 0 0 vx vy)))
    (and
     (> speed *bait-direction-speed*)
     (or
      (angles-within target current *bait-direction-angle*)
      (angles-within (angle-difference target 180) current *bait-direction-angle*)))))

(defun aim-good (aim orientation aim-delta)
  (angles-within aim orientation aim-delta))

(defun drifting-away (x y vx vy orientation)
  (or (and (clockwisep (angle 0 0 (+ x vx) (+ y vy)) (angle 0 0 x y))
	   (counterclockwisep (angle 0 0 x y) orientation))
      (and (counterclockwisep (angle 0 0 (+ x vx) (+ y vy)) (angle 0 0 x y))
	   (clockwisep (angle 0 0 x y) orientation))))

;; A fuction to update state in imaginal for the purposes of providing feedback to tracker
(defun update-slot (slot)
  (schedule-mod-buffer-chunk 'imaginal (list slot 1) 0))

;; Populate game-state buffer
(defun game-state-features-coop (data)
  (if (equal (getf data :state) "score")
      (progn
        (if  (not (equal (chunk-slot-value-fct (buffer-read 'game-state) 'status) 'game-over))
             (push (getf data :POINTS) *game-points*))
        (mod-buffer-chunk 'game-state
                          (list
			   'game nil
			   'status 'game-over
                           'ship-alive nil
                           'x nil
                           'y nil
                           'vx nil
                           'vy nil
			   'speed nil
			   'flight-direction nil
			   'location nil
			   'orientation nil
			   'aim-at-fortress nil
			   'fortress-angle nil
                           'other-x nil
                           'other-y nil
			   'other-vx nil
			   'other-vy nil
			   'other-location nil
                           'shell-x nil
                           'shell-y nil
			   'missile-present nil
                           'fortress-alive nil
                           'fortress-target nil
                           'fortress-back nil )))
      (let* ((gamestate (getf data :GAMESTATE))
             (self-data (if (equal (get-parameter-value :sf-player) 0)
                            (first (getf gamestate :PLAYERS))
                            (second (getf gamestate :PLAYERS))))
             (other-data (if (equal (get-parameter-value :sf-player) 0)
                             (second (getf gamestate :PLAYERS))
                             (first (getf gamestate :PLAYERS))))
             (missiles (getf gamestate :Missiles))
             (shells (getf gamestate :SHELLS))
             (shell-x (getf (getf (first shells) :POS) :X))
             (shell-y (getf (getf (first shells) :POS) :Y))
             (fortress-data (getf gamestate :FORTRESS))
             (fortress-alive (if (getf fortress-data :alive) 'yes 'no))
             (x (getf (getf self-data :POS) :X))
             (y (getf (getf self-data :POS) :Y))
             (vx (getf (getf self-data :VEL) :X))
             (vy (getf (getf self-data :VEL) :Y))
             (orientation (getf self-data :ANGLE))
             (speed (vector-length (list vx vy)))
             (other-x (getf (getf other-data :POS) :X))
             (other-y (getf (getf other-data :POS) :Y))
             (other-vx (getf (getf other-data :VEL) :X))
             (other-vy (getf (getf other-data :VEL) :Y))
             (flight-direction (if (< speed 0.01) orientation (angle 0 0 vx vy))) ;; if stopped, FD is same as orientation
             (fortress-angle (getf fortress-data :ANGLE))
             (self-alive (if (getf self-data :ALIVE) 'yes 'no))
             ;(other-alive (if (getf other-data :ALIVE) 'yes 'no))
	     (fortress-target-self (if (equal (get-parameter-value :sf-player) '0)
                                       (if (equal (getf fortress-data :TARGET) 0) 'yes 'no)
                                       (if (equal (getf fortress-data :TARGET) 1) 'yes 'no)))
	     (fortress-target-other (if (equal (get-parameter-value :sf-player) '0)
                                        (if (equal (getf fortress-data :TARGET) 1) 'yes 'no)
                               (if (equal (getf fortress-data :TARGET) 0) 'yes 'no)))
             ;switch shooter role here
             ; This is new servers v289+ games
	     (role (if (zerop (get-parameter-value :sf-player)) 'shooter 'bait)))
             ; this is older ones
             ;(role (if (zerop (get-parameter-value :sf-player)) 'bait 'shooter)))
          (cond ((getf self-data ':ALIVE)
		 (mod-buffer-chunk 'game-state
                     (list
                      'role role
				    'game 'coop
				    'status 'play
				    'ship-alive self-alive
				    'x x
				    'y y
				    'vx vx
				    'vy vy
				    'speed speed
				    'flight-direction flight-direction
				    'location (location x y fortress-angle fortress-alive)
				    'orientation orientation
				    'aim-at-fortress (angle-difference (angle 0 0 x y) 180)
				    'fortress-angle fortress-angle
				    'other-x other-x
				    'other-y other-y
				    'other-vx other-vx
				    'other-vy other-vy
				    'other-location (location other-x other-y fortress-angle fortress-alive)
				    'shell-x (if shells shell-x 1000)
				    'shell-y (if shells shell-y 1000)
				    'missile-present (if missiles 'yes 'no)
				    'fortress-alive fortress-alive
				    'fortress-target (cond ((equalp fortress-target-self 'yes) 'self)
							   ((equalp fortress-target-other 'yes) 'other)
							   (t 'none))
				    'fortress-back (if (getf fortress-data :VULNERABLE) 'vulnerable 'notvulnerable)
                                    )))
		(t
		 (mod-buffer-chunk 'game-state
                     (list
                      'role role
				    'game nil
				    'status nil
				    'ship-alive 'no
				    'x nil
				    'y nil
				    'vx nil
				    'vy nil
				    'speed nil
				    'flight-direction nil
				    'location nil
				    'orientation nil
				    'aim-at-fortress nil
				    'fortress-angle nil
				    'other-x nil
				    'other-y nil
				    'other-vx nil
				    'other-vy nil
				    'shell-x nil
				    'shell-y nil
				    'missile-present nil
				    'fortress-alive nil
				    'fortress-target nil
				    'fortress-back nil
				    )))))))

(defun update-theta (theta delta)
  ;; Function to learn parameters with stochastic gradient descent
  (- theta (* *sgd-alpha* delta)))

(defun record-features-coop (data)
  (handler-case (progn 
                  
  (game-state-features-coop data)
  (if (not (equal (getf data :state) "score"))
      (let* ((gamestate (getf data :GAMESTATE))
             (points (getf gamestate ':POINTS))
             (self-data (if (equal (get-parameter-value :sf-player) 0)
			    (first (getf gamestate :PLAYERS))
			    (second (getf gamestate :PLAYERS))))
             (self-alive (if (getf self-data :ALIVE) 'yes 'no))
             ;; (x (getf (getf self-data :POS) :X))
             ;; (y (getf (getf self-data :POS) :Y))
             (fortress-data (getf gamestate :FORTRESS))
             (fortress-alive (if (getf fortress-data :alive) 'yes 'no))
             ;; (fortress-angle (getf fortress-data :ANGLE))
             (shells (getf gamestate :SHELLS))
	     (missiles (getf gamestate :MISSILES))
	     (fortress-target-self (if (equal (get-parameter-value :sf-player) '0)
                                       (if (equal (getf fortress-data :TARGET) 0) 'yes 'no)
                                       (if (equal (getf fortress-data :TARGET) 1) 'yes 'no)))
	     ) 
	(if (and (< (length (gethash (current-model) *shells*)) (length shells))
		 (equal fortress-target-self 'yes))
	    (update-slot 'shell-fired))
	; (if (< (length (gethash (current-model) *missiles*)) (length missiles)) (progn
	; 					       (update-slot 'missile-fired)
	; 					       (format t "~a" "MISSILE FIRED"))
	;     )
	(if (and (equal self-alive 'no) (equal (gethash (current-model) *self-alive*) 'yes)) (update-slot 'death))
	(if (and (equal fortress-alive 'no) (equal (gethash (current-model) *fortress-alive*) 'yes)) (update-slot 'kill))
	(if (and (eq fortress-target-self 'no) (< (length missiles) (length (gethash (current-model)  *missiles*))) (equalp fortress-alive 'yes)) (update-slot 'miss))
	(setf (gethash (current-model) *self-alive*) self-alive)
	(setf (gethash (current-model) *fortress-alive*) fortress-alive)
	(setf (gethash (current-model) *shells*) shells)
        (setf (gethash (current-model) *missiles*) missiles)
	(setf *points* points)
        ))
                  )

   ((or error condition)(x) (declare (ignore x)) nil )))

(defun get-server-port ()
  (let ((v (if (and (boundp '*sfcoop-headless-server-port*)
                    (symbol-value '*sfcoop-headless-server-port*))
               (symbol-value '*sfcoop-headless-server-port*)
               3000)))
    (format t "; server port is ~S~%" v)
    v))
  
(defun report-progress (c limit)
  (when (and (boundp '*cluster-progress-update-function*)
             (symbol-value '*cluster-progress-update-function*))
    (funcall (symbol-value '*cluster-progress-update-function*) (/ c limit))))


;;; Redefine this so t -> true and nil -> false in the json encodings

(defun sexp2json (sexp)
  "Convert a simple s-expression into a JSON representation in a string ."
  (cond
   ((eq sexp t)
    '|true|)
   ((eq sexp nil)
    '|false|)
   ((and (listp sexp)
          (evenp (length sexp))
          (loop for rest = sexp then (cddr rest)
                for k = (first rest)
                while rest
                always (keywordp k)))
     (format nil "{~{~A:~A~^,~}}" (mapcar #'sexp2json sexp)))
    ((listp sexp)
     (format nil "[~{~A~^,~}]" (mapcar #'sexp2json sexp)))
    ((symbolp sexp)
     (with-output-to-string (out)
       (write-char #\" out)
       (loop with s = (symbol-name sexp)
             with capitalize = nil
             for i below (length s)
             do (cond
                  ((char= (elt s i) #\-)
                   (setf capitalize t))
                  (capitalize
                   (write-char (char-upcase (elt s i)) out)
                   (setf capitalize nil))
                  (t
                   (write-char (char-downcase (elt s i)) out))))
       (write-char #\" out)))
    ((stringp sexp) (format nil "~S" sexp))
    ((numberp sexp) (format nil "~A" sexp))
    (t (warn "Dunno how to translate ~S to JSON" sexp)
       "null")))


;; redefine this here to add some additional config settings
;; should move this back to the module code ...

(defun connect-to-coop-game (module draw stepping game bot)
  (labels ((bad-connection-exit (message)
             (print-warning message)
             (when (usocket::usocket-p (socket module))
               (usocket:socket-close (socket module))
               (setf (socket module) nil))
                                
             (setf (connection-state module) 'disconnected)
             (return-from connect-to-coop-game nil))
           
           (parse-data-line (message)
             (let ((data (server-read-message module)))
               (if data
                   data
                   (bad-connection-exit message))))
           
           (send-command (cmd error-message)
             (server-send-command cmd module)
             (let ((data (parse-data-line error-message)))
               (unless (and (equal (get-value data :msgtype) "response")
                            (equal (get-value data :cmd) (getf cmd :cmd))
                            (get-value data :success))
                 (bad-connection-exit (format nil "Did not get a successful response for command ~S: ~S" cmd data)))
               data))
           
           (set-config-parameter (key value)
             (send-command (list :cmd key :value value)
                           (format nil "Failed reading response to ~A setting" key))))
    
    (bt:with-recursive-lock-held ((comm-lock module))
      
      (if (eq (connection-state module) 'disconnected)
          
          (let ((socket (handler-case (usocket:socket-connect (sf-host module) (sf-port module) :nodelay :if-supported)
                          (error (x) (bad-connection-exit
                                      (format nil "Error ~/print-error-message/ occurred while trying to open a socket connection to host ~s port ~s" 
                                        x (sf-host module) (sf-port module)))))))
            
            (setf (socket module) socket)
            (setf (connection-state module) 'connecting)
            
            (let ((data (parse-data-line "Failed to read initial config")))
              
              (bt:with-recursive-lock-held ((state-lock module))
              
                ;; create the update times to to total a second
                (setf (game-freq module) (get-value data :fps))
                (setf (update-times module) (create-times (game-freq module)))
                
                (when (not (and (equal (get-value data :msgtype) "stateChange")
                                (equal (get-value data :state) "config")))
                  (bad-connection-exit (format nil "Expected initial config line but got ~s" data)))
                
                (set-config-parameter "drawing" (if draw 1 0))
                (set-config-parameter "player" (server-player module))
                (set-config-parameter "stepping" (ecase stepping (:model "model") (:timer "timer")))
                (when game
		  (set-config-parameter "condition" game))
		(when bot
		  (set-config-parameter "bot" 1))
		(setf (server-stepping module) stepping)
                (setf (name module) (get-model-name))
                (set-config-parameter "id" (name module))

		;(send-command (list :cmd "config" :keys "fortress.cooldownTicks" :value 120) "Failed to set cooldown")
		;(send-command (list :cmd "config" :keys "fortress.smallhexRadius" :value 45) "Failed to set missile penalty")
		;(send-command (list :cmd "config" :keys "game.fortressHitMissilePenalty" :value t) "Failed to set missile penalty")
		;(send-command (list :cmd "config" :keys "shell.ballCoopCollisions" :value t) "Failed to set hexradius")
		;(send-command (list :cmd "config" :keys "missile.ballCoopCollisions" :value t) "Failed to set hexradius")
		;(send-command (list :cmd "config" :keys "player.ballCoopCollisions" :value t) "Failed to set hexradius")
		;(send-command (list :cmd "config" :keys "player.startingDistance" :value 250) "Failed to set hexradius")
                (send-command (list :cmd "continue") "Failed to progress from config screen")
                ;; The next message will be a state change into the
                ;; "game" state. So consider the module connected.
                (setf (connection-state module) 'connected
                  (server-mode module) 'ready)
                t)))
        (print-warning "Model ~s is already connected cannot connect." (current-model))))))


(defun update-game-count ()
  (bt:with-lock-held ((game-counter-lock *game-counter*))
    (pushnew (current-model) (game-counter-updated-players *game-counter*))
    (when (= (length (game-counter-updated-players *game-counter*))
             (game-counter-players *game-counter*))
      (incf (game-counter-current-game *game-counter*))
      (setf (game-counter-updated-players *game-counter*) nil))))
