;;;  -*- mode: LISP; Syntax: COMMON-LISP;  Base: 10 -*-;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 
;;; Author      : Dan Bothell & Shawn Betts
;;; Copyright   : (c) 2018 Dan Bothell
;;; Availability: Covered by the GNU LGPL, see LGPL.txt
;;; Address     : Department of Psychology 
;;;             : Carnegie Mellon University
;;;             : Pittsburgh, PA 15213-3890
;;;             : db30@andrew.cmu.edu
;;; 
;;; 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 
;;; Filename    : sfcoop-module-plus.lisp
;;; Version     : 11.0a1
;;; 
;;; Description : Module for connecting to either the co-op version of
;;;             : Space Fortress through a socket connection, and supporting multiple
;;;             : models playing simultaneously or the original Space Fortress/Track
;;;             : task.
;;;
;;; Bugs        : 
;;;
;;; To do       : [ ] Investigate the chunk table settings on performance.
;;;
;;; 
;;; ----- History -----
;;;
;;; 2018.09.21 Shawn [1.0]
;;;             : * Initial version based on previous sf-module code (which was
;;;                 built on an interface written for Unreal Tournament).
;;; 2018.09.25 Dan
;;;             : * Added the header (stripped the old sf history info).
;;;             : * Cleaned up some of the old comments.
;;;             : * Changed to version 7.5 as required for now.
;;;             : * Just load all .lisp files from ACT-R:space-fortress-modules.
;;;             : * Reworked the whole visual feature system to just provide 
;;;             :   constant interface items (my-ship, my-explosion, other-ship, 
;;;             :   other-explosion, fortresss, fortress-explosion, and score), 
;;;             :   dynamic items (shells and missles), and user items (anything
;;;             :   else) without the need for a "visible" slot to determine if 
;;;             :   they're available.  
;;;             : * Also removed all the funny business with proc-display at 
;;;             :   attention time.
;;;             : * For now, putting all the features into both the loc and obj
;;;             :   chunks.
;;;             : * Removed all of the hooks except the data hook.
;;; 2018.09.26 Dan
;;;             : * Added parameters for print-comm-data and the time to wait
;;;             :   before sending the continue automatically between games.
;;;             : * Save the data records in order (push-last instead of push).
;;;             : * Updated play-sf and play-sf-games to work in a multi-model
;;;             :   situation.  Requires a player to set the :sf-player-number
;;;             :   parameter to enable playing.
;;;             : * Set the server-mode to a symbol instead of string for easier
;;;             :   use in queries (and then use case instead of cond).
;;;             : * Added the update-time slot to the module which will get the
;;;             :   interval length from the game.
;;;             : * Don't queue up the responses -- send them immediately.
;;; 2018.09.27 Dan
;;;             : * Realized there's an issue with synchronizing things at the
;;;             :   start and between games.  Can't just "wait" for the game
;;;             :   start signal because the model will fall behind the real
;;;             :   time clock so it needs to continue to advance the model time
;;;             :   too.  The "right" solution would be to get access to the
;;;             :   real time clock so it could wait and then schedule the
;;;             :   update/change to happen at the current real time, but that
;;;             :   will require a custom clock function and gets ugly (at least
;;;             :   for 7.5).  So, for now just putting in a dummy check event
;;;             :   that'll just burn through time at a fourth the game tick size
;;;             :   (so it isn't wasting too much time spinning but should
;;;             :   still keep the model withing a tick of the real start).
;;;             : * Added a scene-change flag because only need to schedule
;;;             :   proc-display if something updated for efficiency.
;;; 2018.09.28 Dan [1.1]
;;;             : * Create a real schedule of update times based on the fps value.
;;;             : * In model mode can't wait for the step response because 
;;;             :   multiple models are going to send step sequentially.
;;;             : * Fixed some issues with respect to sending the step and 
;;;             :   continue -- it only needs to send a single continue
;;;             :   while waiting for a new game to start.
;;; 2018.10.01 Dan
;;;             : * Working on updating the visual features for the model.
;;; 2018.10.02 Dan [1.2]
;;;             : * Seems stable now and visual features seem to be working.
;;; 2018.10.03 Dan
;;;             : * Fixed the reason slot of ship explosions since the event
;;;             :   only comes over on the initial death message.
;;;             : * Added the other features to the fortress chunks.
;;;             :   [Translate the 1/2 value for target to self/other but that
;;;             :    may change to 0/1 in a future update.]
;;;             : * Fixed the missing id slot in shots and missiles.
;;;             : * Fixed a bug with remove-visual-feature.
;;;             : * Updated for 0.53 server that sends 0/1 for fortress targets.
;;; 2018.10.04 Dan [1.3]
;;;             : * Use require-compiled to load the extensions and move them
;;;             :   into a subdirectory under space-fortress-modules.
;;; 2018.10.18 Dan [2.0]
;;;             : * Start converting this over to 7.12.
;;;             :   - Use usocket instead of uni-files.
;;;             :   - Use built-in json encoder/decoder.
;;;             :   - Don't use the module as a device.
;;;             :   - Add locks around the module code.
;;;             :   - Use visicon-features instead of chunks.
;;;             :   - Keep it single-threaded right now, but eventually will
;;;             :     switch to having a second thread parse the incoming data,
;;;             :     and in timer mode just schedule the updates as they come
;;;             :     in (model mode should still use the fixed update events).
;;;             : * Simplify the game counting so it doesn't have to iterate
;;;             :   over all the connected models.
;;; 2018.10.19 Dan
;;;             : * CLJSON is slow -- half the run time for the game is spent
;;;             :   in the decoder!  Going back to Shawn's simple JSON parsers.
;;; 2018.10.22 Dan
;;;             : * Add the fortress-explosion and fortress chunks.
;;;             : * Still using the repeated checks for start, but once that's
;;;             :   in a background process it could just schedule for "now".
;;; 2018.11.05 Dan
;;;             : * Added a chunk named bonus.
;;;             : * Ignore the world parameter in create/update player-visual-feature
;;;             :   (should probably just remove it).
;;; 2020.01.28 Dan
;;;             : * Fixed the call to set-default-vis-loc-slots since it's a 
;;;             :   macro as of 7.13.2.
;;; 2020.10.14 Dan [3.0]
;;;             : * Schedule all the events with a priority of :min so that they
;;;             :   will cycle among models if they're all scheduling events for
;;;             :   the same time repeatedly.
;;;             : * Don't schedule the next update for a model immediately upon
;;;             :   the game starting.  Instead, wait for all the models to be
;;;             :   connected and then schedule them all together.
;;;             : * Make sure that handle-server-message always gets called and
;;;             :   have it clear pending-step for continue messages too.
;;; 2022.01.24 Dan [11.0a1]
;;;             : * First pass at combining the coop and older interfaces.
;;;             : * Incremented the version number to the sum of the versions
;;;             :   of the two being combined.
;;;             : * Taking a brute force approach of just having a flag set in
;;;             :   the module when starting a game that's tested in all the
;;;             :   interface code to pick the right version.
;;;             : * The game running functions will set that and handle the 
;;;             :   reconnection if continuing on a different server.
;;;             : * Dropping the visual information for now to keep things 
;;;             :   simpler since people aren't using it.
;;;             : * There are separate data-hook parameters for the interfaces:
;;;             :   :sf-data-hook and :coop-data-hook.
;;;             : * The play-sf and play-sf-games functions run the older task
;;;             :   and play-sf-coop and play-sf-coop-games run the coop task.
;;;             :   All of those functions have a new optional parameter of 
;;;             :   :port which can be specified with the appropriate tcp/ip
;;;             :   port of the server.  The assumption being the servers will
;;;             :   have the same address, but use different ports.  When a
;;;             :   run is continued using a different interface it should 
;;;             :   cleanly disconnect the old game and connect to the new one.
;;; 2022.02.02 Dan
;;;             : * Fixed some cut&paste and search and replace issues.
;;;             : * In the older game running functions loop over the models so
;;;             :   that multiple can be playing at the same time, and change
;;;             :   the game count testing function to only stop when all models
;;;             :   have completed the games.
;;; 2022.02.04 Dan
;;;             : * Changed the update-coop-world event to not be maintenance
;;;             :   now since there's no proc-display to trigger conflict-
;;;             :   resolution to start again.
;;; 2022.02.11 Dan
;;;             : * The old game server never actually used the port setting so
;;;             :   always connected to 3000.  Now it uses the host and port 
;;;             :   values.
;;;             : * It was calling reset multiple times when I added the loop
;;;             :   over players...
;;; 2022.06.06 Dan
;;;             : * Added a flag and lock for continuing because if they don't
;;;             :   end the game at the same time they can get into a race 
;;;             :   condition with the send-continue actions -- if one schedules
;;;             :   it and follows up with a wait-for-game before the other
;;;             :   sends it's stuck with the clock not advancing.
;;; 2022.06.08 Dan
;;;             : * Another adjustment to send-continue because it seems like
;;;             :   when there's extra responses (unexpected keypresses) it can
;;;             :   get out of sync, and when scheduling the next update-coop-
;;;             :   world after a wait-for-game-start need to make sure that is
;;;             :   :max otherwise the other model could prevent it from ever
;;;             :   happening and freeze the system.
;;; 2022.06.09 Dan
;;;             : * That's still not quite right, so trying a 0 delay for
;;;             :   the event as well to try and keep them together.
;;;             : * Also adding the time to the event printouts so I can see the
;;;             :   ordering of things.
;;; 2022.06.14 Dan
;;;             : * Added a :game-ticks parameter because it's possible for
;;;             :   there to be a delay between the last game tick and the
;;;             :   state change that results in the model delaying an extra
;;;             :   tick which affects the determinism of the system.  Now,
;;;             :   if the tick count matches game-ticks it schedules the
;;;             :   update with no delay so it catches the game state at the
;;;             :   "right" time.
;;; 2022.06.17 Dan
;;;             : * Updated json2sexp so that it can handle an empty object and
;;;             :   consider it as nil.
;;; 2022.09.16 Dan
;;;             : * Added another keyword parameter to play-sf-coop :game which
;;;             :   if set is passed as the condition in setup of the game, and
;;;             :   the value "dynamic-ships" can be used to run the older game.
;;; 2022.09.23 Dan
;;;             : * Added another keyword parameter to play-sf-coop :bot which
;;;             :   if set to non-nil will enable the bait bot in the server.
;;;             :   Assumes only one playing model defined (a shooter).
;;; 2022.09.28 Dan
;;;             : * Added a check to the score handling so that if there's only
;;;             :   one model it always sends the continue immediately.
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; The code
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;; Standard ACT-R checks for putting things in an appropriate package as needed.

#+:packaged-actr (in-package :act-r)
#+(and :clean-actr (not :packaged-actr) :ALLEGRO-IDE) (in-package :cg-user)
#-(or (not :clean-actr) :packaged-actr :ALLEGRO-IDE) (in-package :cl-user)

;;; Do a version check.

(written-for-act-r-version "7.21" (format nil "Space Fortress module.~%           Needs fully updated extended-motor-actions."))

;;; Load the necessary ACT-R extensions.  Some are included with ACT-R and others
;;; are specific to the SF interface and included separately.
;;; Also create a logical host called SF-DATA pointing to the directory where
;;; this file is located.

(require-extra "extended-motor-actions")

;; require the loader which will load all of the .lisp files from the 
;; space-fortress-modules/load directory.

(require-compiled "SF-MODULES" "ACT-R:space-fortress-modules;loader.lisp")

;; Create the logical SF-DATA to point to the directory of this file.

(setf (logical-pathname-translations "SF-DATA")
  `(("**;*.*" ,(namestring (merge-pathnames "**/*.*" *load-truename*)))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Set the system parameters to adjust the chunk table initial size.  Improved
;;; performance in original SF due to the number of chunks created while playing
;;; (mostly visual and imaginal).
;;; May or may not be necessary now. 

(ssp :mcts 45000 :mctrt .75)

;;; Simple json parser for game communication in coop

(defun sexp2json (sexp)
  "Convert a simple s-expression into a JSON representation in a string ."
  (cond
    ((and (listp sexp)
          (evenp (length sexp))
          (loop for rest = sexp then (cddr rest)
                for k = (first rest)
                while rest
                always (keywordp k)))
     (format nil "{~{~A:~A~^,~}}" (mapcar #'sexp2json sexp)))
    ((listp sexp)
     (format nil "[~{~A~^,~}]" (mapcar #'sexp2json sexp)))
    ((symbolp sexp)
     (with-output-to-string (out)
       (write-char #\" out)
       (loop with s = (symbol-name sexp)
             with capitalize = nil
             for i below (length s)
             do (cond
                  ((char= (elt s i) #\-)
                   (setf capitalize t))
                  (capitalize
                   (write-char (char-upcase (elt s i)) out)
                   (setf capitalize nil))
                  (t
                   (write-char (char-downcase (elt s i)) out))))
       (write-char #\" out)))
    ((stringp sexp) (format nil "~S" sexp))
    ((numberp sexp) (format nil "~A" sexp))
    (t (warn "Dunno how to translate ~S to JSON" sexp)
       "null")))

(defun json2sexp (str &optional ( start 0))
  "A json parser. Not complete but good enough for communicating with
the server."
  (let ((ch (char str start)))
    (cond
      ((char= ch #\{)
       (incf start)
       (if (char= (char str start) #\})
           (values nil (incf start))
         (loop with k = nil
             with v = nil
             with ret = nil
             with done = nil
             until done
             do (multiple-value-setq (k start) (json2sexp str start))
                (when (char/= (char str start) #\:) (error "dict parse error"))
                (incf start)
                (multiple-value-setq (v start) (json2sexp str start))
                (push (intern (string-upcase k) :keyword) ret)
                (push v ret)
                (if (char= (char str start) #\,)
                    (incf start)
                    (progn
                      (when (char/= (char str start) #\}) (error "dict parse error"))
                      (incf start)
                      (setf done t)))
             finally (return (values (nreverse ret) start)))))
      ((char= ch #\[)
       (incf start)
       (when (char= (char str start) #\])
         (incf start)
         (return-from json2sexp (values nil start)))
       (loop with v = nil
             with ret = nil
             with done = nil
             until done
             do (multiple-value-setq (v start) (json2sexp str start))
                (push v ret)
                (if (char= (char str start) #\,)
                    (incf start)
                    (progn
                      (when (char/= (char str start) #\]) (error "array parse error"))
                      (incf start)
                      (setf done t)))
             finally (return (values (nreverse ret) start))))
      ((char= ch #\")
       (incf start)
       (loop for i from start below (length str)
             when (char= (char str i) #\")
               do (return (values (subseq str start i) (1+ i)))
             finally (error "string parse error")))
      ((or (and (char>= ch #\0)
                (char<= ch #\9))
           (char= ch #\-))
       (loop with beg = (if (char= ch #\-) (1+ start) start)
             with dot = nil
             with exponent = nil
             with done = nil
             for i from beg below (length str)
             until done
             do (cond
                  ((char= (char str i) #\.)
                   (if (or dot exponent)
                       (error "number parse error")
                       (setf dot t)))
                  ((char= (char str i) #\e)
                   (if exponent
                       (error "exponent parse error")
                       (setf exponent t)))
                  ((char= (char str i) #\-)
                   (unless (char= (char str (1- i)) #\e)
                     (error "exponent minus parse error")))
                  ((and (char>= (char str i) #\0)
                        (char<= (char str i) #\9)))
                  (t (setf done t)))
             finally (return (values (read-from-string (subseq str start (1- i)) t) (1- i)))))
      ((string= str "true" :start1 start :end1 (+ start 4))
       (values t (+ start 4)))
      ((string= str "false" :start1 start :end1 (+ start 5))
       (values nil (+ start 5)))
      ((string= str "null" :start1 start :end1 (+ start 4))
       (values nil (+ start 4)))
      (t (error "top-level parse error")))))



;;; Communication to the game in old versions is via text based commands on one line.
;;; The sending functions are responsible for proper formatting.

(define-constant *cr/lf* (format nil "~C~C" #\cr #\lf))

(defmacro server-write-line (world string &rest args)
  "Write a CRLF-terminated line"
  `(if (or (null (server ,world))
           (not (sfapi-connected-p (server ,world))))
       (print-warning "Can't send command {~S} because connection closed" 
                   (format nil ,string ,@args))
     (let ((string-to-write (format nil ,string ,@args)))
       (when (record-outgoing ,world)
         (push (list :out (mp-time) (get-internal-real-time) string-to-write)
               (data-record ,world)))

       (sfapi-write (server ,world) 
                    (concatenate 'string string-to-write *cr/lf*)))))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Structure to keep track of models playing, games played, and syncing game starts

(defstruct game-counter ready-players players current-game limit updated-players (lock (bt:make-lock "game-counter")))
(defvar *game-counter* (make-game-counter))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Need to synchronize the continues assumes 2 players, but could be changed to
;;; handle more.

(defstruct continue-check delayed time (lock (bt:make-lock "sync")))
(defvar *continue* (make-continue-check))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Create a class which will be a module for each model.

(defclass sf-plus ()
  (;; The game type flag either :coop or :older
   (game-type :initform nil :accessor game-type)
     
   ;; shared connection info
   (host :initform nil :accessor sf-host)
   (port :initform nil :accessor sf-port)
   
   ;; coop connection info
   (socket :initform nil :accessor socket)
   (server-mode :initform nil :accessor server-mode)
   (connection-state :initform nil :accessor connection-state)
   (server-player :initform nil :accessor server-player)
   (server-stepping :initform nil :accessor server-stepping)
   
   ;; non-coop connection info
   (server-type :initform nil :accessor sf-server-type)
   (server :initform nil :accessor server)
   (server-current-screen :initform nil :accessor server-current-screen)
   (game-state :initform nil :accessor game-state)
   
      
   ;; communication logging
   (collect-incoming :initform nil :accessor collect-incoming)
   (record-outgoing :initform nil :accessor record-outgoing)
   (data-record :initform nil :accessor data-record)
   (print-communication :initform t :accessor print-communication)
   
   
   ;; user hook parameters
   (data-hook :initform nil :accessor data-hook)
   (coop-data-hook :initform nil :accessor coop-data-hook)
   
   ;; track some game stats as it plays (both use it)
   (game-stats :initform (list (list 0 0)) :accessor game-stats)
   ;; stats from non-coop
   (ship-alive :initform nil :accessor ship-alive)
   (fortress-killed :initform nil :accessor fortress-killed)
   
   
   ;; connection and running control
   ;;; common 
   (name :initform nil :accessor name)
   (change-drawing-mode :initform nil :accessor change-drawing-mode)
   
   ;;; coop
   (pending-step :initform nil :accessor pending-step)
   (continue-sent :initform nil :accessor continue-sent)
   (continue-time :initform 5 :accessor continue-time)
   (update-time :initform 0 :accessor update-time)
   (update-times :initform '(1000) :accessor update-times)
   (game-freq :initform 1 :accessor game-freq)
   
   ;;; non-coop
   
   (last-time :initform 0 :accessor last-time)
   (game-count :initform 0 :accessor game-count)
   (game-limit :initform -1 :accessor game-limit)
   (key-handlers :initform nil :accessor key-handlers)
   (game-condition :initform nil :accessor game-condition)
   (responses :initform nil :accessor responses)
   (auto-response :initform nil :accessor auto-response)
   (fort-exists :initform nil :accessor fort-exists)
   (fortress-there :initform nil :accessor fortress-there)
   
   (game-ticks :initform 10800 :accessor game-ticks)
   
   ;; locks for various pieces
   (comm-lock :initform (bt:make-recursive-lock "sf-comm") :accessor comm-lock)
   ; socket server-mode connection-state data-record server-stepping
   
   (param-lock :initform (bt:make-recursive-lock "sf-param") :accessor param-lock)
   ; host port server-player collect-incoming record-outgoing data-hook continue-time print-communication
   
   (state-lock :initform (bt:make-recursive-lock "sf-state") :accessor state-lock)
   ; everything else...
   ))

;;; The older interface key mapping.
;;;
;;; To keep things simple it's just assuming the hand positions are over
;;; the correct keys.  The model may set them that way if it wants, but
;;; it will not affect how it plays.
;;;
;;; [2.0] These are now mapped to the keycodes needed instead of the key names.
;;; with the current actions in the config being:  a,w,d for movement (left ring, middle, index respectively)
;;; space for fire (right index) and j for IFF (right middle)
;;; all others are mapped to z as a default.
;;;
;;; [3.0] The model presses are now reported as keys so need to change the
;;; mapping appropriately with the assumption that the models hands will
;;; just be on the home row.  Thus s,d,f are left, thrust, right and
;;; j,k,l are fire, iff, and bonus.
  
  
(defun old-model-key-to-sf-key (key)
  (case key
    (#\d 119)
    (#\s  97)
    (#\f 100)
    (#\j  32)
    (#\k 106)
    (#\l 107)
    (t  122)))


;;; The coop interface 

;;; Since the model actions are generated based on hand and finger specifications
;;; just assume the default hand positions (home row) to keep things simple since
;;; hand position doesn't affect the punch timing (there's no hand/finger movement
;;; necessary to play).
;;;

(defun coop-model-key-to-sf-key (key)
  (cond ((string-equal key "d")
         "thrust")
        ((string-equal key "s")
         "left")
        ((string-equal key "f")
         "right")
        ((string-equal key "j")
         "fire")
        (t "nothing")))


;; The monitoring functions

(defun model-keypress (model key)
  (declare (ignore model))
  (let ((world (get-module :sf-plus)))
    (if world
      (bt:with-recursive-lock-held ((state-lock world))
        (cond ((eq (game-type world) :older)
               (push-last (cons 'keydown (old-model-key-to-sf-key (char key 0))) (responses world)))
              ((eq (game-type world) :coop)
               (server-send-command (list :cmd "keydown" :value (coop-model-key-to-sf-key key)) world))
              (t
               (print-warning "Keypress generated without a valid game-type ~s" (game-type world)))))
      (print-warning "Keypress generated without a current module"))))

(defun model-keyrelease (model key)
  (declare (ignore model))
  (let ((world (get-module :sf-plus)))
    (if world
      (bt:with-recursive-lock-held ((state-lock world))
        (cond ((eq (game-type world) :older)
               (push-last (cons 'keyup (old-model-key-to-sf-key (char key 0))) (responses world)))
              ((eq (game-type world) :coop)
               (server-send-command (list :cmd "keyup" :value (coop-model-key-to-sf-key key)) world))
              (t
               (print-warning "Keyrelease generated without a valid game-type ~s" (game-type world)))))
      (print-warning "Keyrelease generated without a current module"))))

(add-act-r-command "sf-keypress" 'model-keypress "Monitor for output-key actions for space fortress module.")
(add-act-r-command "sf-keyrelease" 'model-keyrelease "Monitor for release-key actions for space fortress module.")

(monitor-act-r-command "output-key" "sf-keypress")
(monitor-act-r-command "release-key" "sf-keyrelease")


;; Functions for use if one wants to do something other than
;; space fortress with current ACT-R session and then return to space
;; fortress afterwards.

(defun remove-sf-monitors ()
  (remove-act-r-command-monitor "output-key" "sf-keypress")
  (remove-act-r-command-monitor "release-key" "sf-keyrelease"))

(defun restore-sf-monitors ()
  (monitor-act-r-command "output-key" "sf-keypress")
  (monitor-act-r-command "release-key" "sf-keyrelease"))


;;; older interface 

;;; Send the model's actions and any automatic responses that
;;; are generated for some of the between game screens used in
;;; the fMRI task.

(defun send-all-pending-model-actions (world)
  
  (bt:with-recursive-lock-held ((state-lock world))
    (dolist (x (responses world))
      (server-write-line world "~a ~d 0" (car x) (cdr x)))
  
    (setf (responses world) nil)
    
    (let (new)
      
      (dolist (x (auto-response world))
        (server-write-line world "~a ~d 0" (car x) (cdr x))
        (when (eq (car x) 'keydown)
          (push-last (cons 'keyup (cdr x)) new)))
      
      (setf (auto-response world) new))))
  
;;; Read a line of data from the server skipping blank lines
;;; and terminating the run if there's an error encountered.

(defun get-line-of-data-from-game (world)
  (loop
    (multiple-value-bind (line err)
        (ignore-errors (sfapi-read (server world)))

      (when (subtypep (type-of err) 'condition)
        (bt:with-recursive-lock-held ((state-lock world))
          (setf (game-state world) 'disconnected))
        (print-warning "Error reading from game server.  Cannot continue.")
        (schedule-break-relative 0 :priority :max :details "Game connection error" :time-in-ms t)
        (return nil))
      (unless (string= line "")
        (return line)))))

(defun incf-game-count (world) (incf (game-count world)))

(defun update-world (world)
  
  ;; send everything that's happened since the last update
  (if (eq (server-current-screen world) 'score)
      ;; The score screen does not accept key events
      (setf (responses world) nil
            (auto-response world) nil)
      (send-all-pending-model-actions world))
  
  ;; Always send a continue.
  ;; The next update will be scheduled based on the
  ;; time/delay from the game data that's read here.
  
  (server-write-line world "continue")
  
  (let ((state-line (get-line-of-data-from-game world)))
    
    (unless state-line (return-from update-world))
    
    (let* ((data (read-from-string state-line))
           (delay (if (string-equal (getf data :mode) "delay")
                      (getf data :delay_duration)
                    0)))
      
      ;; record the data and send it to the hook if needed
      
      (when (collect-incoming world)
        (push (list :in (mp-time) (get-internal-real-time) data) (data-record world)))
      
      (awhen (data-hook world) (ignore-errors (funcall it data)))
      
      
      ;; Update the features and time info based on the type of info this is
      
      (setf (server-current-screen world) (read-from-string (getf data :screen-type)))
      
      (case (server-current-screen world)
        
        (progress ;; just ignore it
         )
        
        (basic-task ;; press z to skip it and reset time offset
         (push-last (cons 'keydown 122) (auto-response world))
         (setf (last-time world) 0)
         )
        
        (instructions ;; press z to skip it
         (push-last (cons 'keydown 122) (auto-response world))
         )
        
        (wait-caret ;; press the ^ key to skip it
         (push-last (cons 'keydown 94) (auto-response world))
         )
        
        (fixation ;; assume a 10 second delay if none specified
         (when (zerop delay) 
           (setf delay 10000))
         )
        
        (fmri-task ;; press z to advance and record the visual
                   ;; features for the game type and reset time offset
         
         (push-last (cons 'keydown 122) (auto-response world))
         
         (when (zerop delay) ;; assume 10 second delay if none specified
           (setf delay 10000))
         
         (setf (last-time world) 0))
        
        
        (foe-mines ;; press z to advance and record the visual
                   ;; features for the mine letters 
           
         (push-last (cons 'keydown 122) (auto-response world))
         
         (when (zerop delay) ;; assume 10 second delay if none specified
           (setf delay 10000))
         
         )
        
        (total-score ;; gets sent twice one with a delay and one without
                     ;; in the delay version provide the visual feature
                     ;; so the model sees the end of game and clear some 
                     ;; of the hack flags
                     ;; In the no delay version hit z to skip it, record results,
                     ;; and schedule an event for 1 second from not to note the
                     ;; game ended.
           

         (cond ((zerop delay)
                (push-last (cons 'keydown 122) (auto-response world))
                (push-last (getf data :raw-pnts) (first (game-stats world)))
                (push-last (getf data :bonus) (first (game-stats world)))
                (push-last (getf data :total-bonus) (first (game-stats world)))
                (push-last (append (getf data :base-stats) (getf data :event-stats) (getf data :input-stats)) (first (game-stats world)))
                
                (push (list 0 0) (game-stats world))
                (schedule-event-relative 1000 'incf-game-count :params (list world) :priority :max :time-in-ms t :maintenance t :output nil)
                (setf (last-time world) 0))
               
               (t
                )))
        
        
        (score ;; there's only one score message for the current game servers
               ;; when using model_interface 2 which is all this will handle now
               ;; this might cause problems if running with an older space fortress
               ;; game server.
               ;; 
               ;; Show the trial-end feature to the model, record the
               ;; game stats, and schedule an event to mark the game
               ;; as complete after 1 second.
           
         
         
         ;; record the values from this game and set the list for the next game
         
         (push-last (getf data :raw-pnts) (first (game-stats world)))
         (push-last (getf data :bonus) (first (game-stats world)))
         (push-last (getf data :total-bonus) (first (game-stats world)))
         (push-last (append (getf data :base-stats) (getf data :event-stats) (getf data :input-stats)) (first (game-stats world)))
         
         (setf (last-time world) 0)
         (setf delay 1000)
         (push (list 0 0) (game-stats world))
         
         (schedule-event-relative delay 'incf-game-count :params (list world) :priority :max :time-in-ms t :maintenance t :output nil))
        
        (bonus ;; just quit the game since it got to the end
               ;; and there's nothing else that can be done
               ;; Not used in the current servers (2016), but left in 
               ;; anyways.
           
         (server-write-line world "quit"))
        
        (config ;; this is now handled in the connection function
                ;; Shouldn't see any while the game is running.
         
         (print-warning "Unexpected config data during the task: ~S" data))
        
        
        (game ;; This is where all the action happens...
           
         ;; If the display is being turned on/off between runs
         ;; send that notice now because it's only valid during
         ;; a game update not config or score
         
         (awhen (change-drawing-mode world)
                (server-write-line world "drawing ~d" it)
                (setf (change-drawing-mode world) nil))
           
         ;; compute how long to delay based on the time provided
         ;; unless there's an explicit delay given
         
         (let ((next-time (getf data :time)))
           (cond ((zerop delay)
                  (setf delay (- next-time (last-time world)))
                  (setf (last-time world) next-time))
                 (t
                  (setf (last-time world) (+ delay (last-time world))))))
         
         ;; Update all the visual features for the model
           ;; gone...
         
           ;; still handle some of the internal values
           (awhen (getf data :fortress)
                  (if (getf it :alive)
                      (when (fort-exists world) 
                        (setf (fortress-killed world) nil)
                        (setf (fortress-there world) t))
                    (setf (fortress-there world) :dead)))
             
           (awhen (getf data :ship) 
                    
                  (if (getf it :alive)
                      (setf (ship-alive world) t)
                    
                    (when (ship-alive world)
                      (incf (first (first (game-stats world))))
                      
                      (setf (ship-alive world) nil))))
             
             )
            
        
        (t (print-warning "unexpected screen-type: ~s" data)))
    
    
      ;; Do some special processing for some of the visual information
      
      
      ;; there should be fortress but it's not visible
      ;; originally there wasn't a fortress feat if it was
      ;; dead, but that may have changed now.
      ;; To be safe still handle this separately and
      ;; not just change the kind on the fortress feature.
      
        (when (eq (fortress-there world) :dead)
          (setf (fortress-there world) t)
          (unless (fortress-killed world)
            (incf (second (first (game-stats world))))
            (setf (fortress-killed world) t)))
      
      ;; Schedule the next update based on the needed delay
      
      (schedule-event-relative delay 'update-world :destination :sf-plus :module :sf-plus :maintenance nil :time-in-ms t))))

;;; coop interface

;;; JSON parser in ACT-R uses conses for JSON objects (in and out)
;;; use these to convert the lists used for output and get data from input items.
;;;
;;; Since I switched back to the simple parser, these aren't really needed
;;; but keeping them around just incase I decide to switch to a different
;;; version again at some point.

(defun get-value (data key)
  (getf data key)) ; (cdr (assoc key data)))

(defun list-to-conses (data) data) 
#|
  (do* ((c (pop data) (pop data))
        (d (pop data) (pop data))
        (i (cons (string-downcase (symbol-name c)) d) (cons (string-downcase (symbol-name c)) d))
        (r (list i) (push-last i r)))
       ((null data) r)))
|#

;;; coop communication parsing

(defun server-send-command (cmd &optional module)
  "Write a command to the server socket."
  (let ((world (or module (get-module :sf-plus))))
    (bt:with-recursive-lock-held ((param-lock world))
      (if (eq (connection-state world) 'disconnected)
          (print-warning "Can't send command {~S} because connection closed" cmd))
      (bt:with-recursive-lock-held ((comm-lock world))
        (when (record-outgoing world)
          (push-last (list :out (mp-time) (get-internal-real-time) cmd)
                     (data-record world)))
        (let ((js (sexp2json (list-to-conses cmd))))
          (when (print-communication world) (model-output ">>>~D ~A~%" (get-internal-real-time) js))
          (format (usocket:socket-stream (socket world)) "~a" js)
          (force-output (usocket:socket-stream (socket world))))))))

(defun server-read-message (world)
  "Read a line of data from the server skipping blank lines and terminating the run if there's an error encountered."
  (loop
    (bt:with-recursive-lock-held ((comm-lock world))
      (multiple-value-bind (line err)
          (ignore-errors (read-line (usocket:socket-stream (socket world))))

        (when (or (null line) (subtypep (type-of err) 'condition))
          (setf (connection-state world) 'disconnected)
          (print-warning "Error reading from game server.  Cannot continue.")
          (schedule-break-relative 0 :priority :max :details (format nil "Game connection error for model ~s" (current-model)) :time-in-ms t)
          (return nil))
        
        (unless (string= line "")
          (let (print collect hook (data (json2sexp line)))
            (bt:with-recursive-lock-held ((param-lock world))
              (setf print (print-communication world))
              (setf collect (collect-incoming world))
              (setf hook (coop-data-hook world)))
            
            (when print (model-output "<<<~d ~A~%" (get-internal-real-time) line))
            
            (when hook
              (multiple-value-bind (res e)
                  (ignore-errors (funcall hook data))
                (declare (ignore res))
                (when (subtypep (type-of e) 'condition)
                  (print-warning "Condition: ~s encountered while calling data-hook ~s" e hook))))
            
            (when collect
              (push-last (list :in (mp-time) (get-internal-real-time) data) (data-record world)))
            
            (return data)))))))


;; List of update lengths for 1 second based on frequency

(defun create-times (freq)
  (multiple-value-bind (base rem) (truncate 1000 freq)
    (let ((times (make-list freq :initial-element base)))
      (unless (zerop rem)
        (let* ((total (/ rem freq))
               (step total)
               (inc 0))
          (dotimes (i freq)
            (when (> (truncate total) inc)
              (incf (nth i times))
              (incf inc))
            (incf total step))))
      times)))

(defun wait-for-game-start (world)
  (if (listen (usocket:socket-stream (socket world)))
    (let ((data (server-read-message world)))
      ;(format t "Wait ~S got ~S~%" (current-model) data)
      (if (and (equal (getf data :msgtype) "response")
               (equal (getf data :cmd) "continue"))
          (progn
            (handle-server-message world data)
            (schedule-next-update world 0))
          (if (and (equal (getf data :msgtype) "stateChange")
                   (equal (getf data :state) "game"))
              (progn 
                (signal-player-ready data))
              (progn
                (print-warning "Unexpected incoming message ~s while waiting for game start." data)
                (schedule-next-update world 0)))))
  ;; if there wasn't any data reschedule to try again at current time
    (schedule-next-update world 0)))


(defun signal-player-ready (data)
  (bt:with-lock-held ((game-counter-lock *game-counter*))
                     (unless (game-counter-ready-players *game-counter*)
                       (format t "Model ~s ready first~%" (current-model)))
    (push (cons (current-model) data) (game-counter-ready-players *game-counter*))
                     (when (= (length (game-counter-ready-players *game-counter*)) 
                              (game-counter-players *game-counter*))
                       (dolist (m (game-counter-ready-players *game-counter*))
                         (with-model-eval (car m)
                           (handle-server-message (get-module :sf-plus) (cdr m))
                           (schedule-next-update (get-module :sf-plus) 0 t)))
                       (setf (game-counter-ready-players *game-counter*) nil))))


(defun handle-server-message (world data)
  (cond
   ((equal (get-value data :msgtype) "response")
    (if (get-value data :success)
        (cond
         ((or (equal (get-value data :cmd) "gameState")
              (equal (get-value data :cmd) "step")
             )
          (when (equal (get-value data :cmd) "step")
            (bt:with-recursive-lock-held ((state-lock world))
              (setf (pending-step world) nil)))
          (handle-state-game world (get-value data :gamestate)))
         ((equal (get-value data :cmd) "continue")
          (bt:with-recursive-lock-held ((state-lock world))
            (setf (pending-step world) nil))
          (schedule-next-update world))
          (t
          ;; ignore for now
          ))
      (cond
       ((equal (get-value data :cmd) "step")
        (bt:with-recursive-lock-held ((state-lock world))
          (setf (pending-step world) nil)))
       (t (print-warning "Response is unsuccessful: ~S" data)))))
   ((equal (get-value data :msgtype) "stateChange")
    (handle-state-change world data))
   ((equal (get-value data :msgtype) "gameStateUpdate")
    (handle-state-game world (get-value data :gamestate)))
   (t
    (print-warning "Unknown message type: ~S" data))))


(defun handle-state-change (world data)
  (let ((state (read-from-string (get-value data :state))))
    (bt:with-recursive-lock-held ((comm-lock world)) (setf (server-mode world) state))
    (case state
      (score
       (handle-state-score world data))
      (game
       (handle-state-game world (get-value data :gamestate)))
      (t
       (print-warning "Unknown state change: ~S" data)))))

(defun handle-state-score (world data)
  (bt:with-recursive-lock-held ((state-lock world))
    (push-last (get-value data :rawPoints) (first (game-stats world)))
    (push-last (get-value data :bonus) (first (game-stats world)))
    (push-last (get-value data :total-bonus) (first (game-stats world)))
  
    (push (list 0 0) (game-stats world))
 
    (update-game-count)
  
    (setf (continue-sent world) nil)
    
    ;; no visual features for now
    ))

(defun handle-state-game (world data)
  
  (let ((player (bt:with-recursive-lock-held ((param-lock world))
                  (server-player world))))
    
    (bt:with-recursive-lock-held ((state-lock world))
      
      ;; increment the time until next update
      (unless (= (get-value data :tick) (game-ticks world))
         (incf (update-time world) (nth (mod (get-value data :tick) (game-freq world)) (update-times world))))
    
      ;; record deaths and fortress kills
      (when (find "kill-player" (get-value data :events)
                  :test #'string=
                  :key (lambda (e) (get-value e :tag)))
        (incf (first (first (game-stats world)))))
      
      (when (find "kill-fortress" (get-value data :events)
                  :test #'string=
                  :key (lambda (e) (get-value e :tag)))
        (incf (second (first (game-stats world)))))
      
      ;; Update players
      ;; removed the visual updating for now...
      )))

(defun send-continue (world &optional skip)
  (unless skip
    (server-send-command '(:cmd "continue") world)
    (setf (continue-sent world) t))
  (let ((data (server-read-message world)))
    (when data (handle-server-message world data))
    (if (and (equal (get-value data :msgtype) "response")
             (equal (get-value data :cmd) "continue"))
        (unless (get-value data :success)
          (print-warning "Continue command failed somehow"))
      (progn 
        (print-warning "Unexpected incoming data ~s when sending a continue" data)
        (send-continue world t)))
    ))


(defun schedule-next-update (world &optional delay force-max)
  (case (bt:with-recursive-lock-held ((comm-lock world)) (server-mode world))
    (ready ;; shouldn't happen now because we wait for the game data
     (schedule-event-relative (or (and (numberp delay) delay)
                                  0)
                              'wait-for-game-start :destination :sf-plus :module :sf-plus :output 'high :time-in-ms t :priority :min))
    (game
     (schedule-event-relative (or (and (numberp delay) delay)
                                  (bt:with-recursive-lock-held ((state-lock world)) (update-time world)))
                              'update-coop-world :destination :sf-plus :module :sf-plus :maintenance nil :time-in-ms t :output 'high :priority (if force-max :max :min)))
    (score
     (bt:with-recursive-lock-held ((state-lock world))
       (if (continue-sent world)
           (schedule-event-now 'wait-for-game-start :destination :sf-plus :module :sf-plus :output 'high :maintenance t :priority :min)
         (bt:with-recursive-lock-held ((continue-check-lock *continue*))
            (cond ((= (length (mp-models)) 1)
                   (let ((continue-time (+ (mp-time-ms)
                                           (or (and (numberp delay) delay)
                                               (bt:with-recursive-lock-held ((param-lock world))(continue-time world))))))
                      (schedule-event continue-time 'send-continue :destination :sf-plus :module :sf-plus 
                                      :maintenance t :time-in-ms t :output 'high :priority :min))
                   

                  (null (continue-check-delayed *continue*))
                   (setf (continue-check-delayed *continue*) (current-model))
                   )
                  ((not (eq (continue-check-delayed *continue*) (current-model)))
                   (let ((continue-time (+ (mp-time-ms)
                                           (or (and (numberp delay) delay)
                                               (bt:with-recursive-lock-held ((param-lock world))(continue-time world))))))
                   
                     (schedule-event continue-time 'send-continue :destination :sf-plus :module :sf-plus :maintenance t :time-in-ms t :output 'high :priority :min)
                     (with-model-eval (continue-check-delayed *continue*) 
                       (schedule-event continue-time 'send-continue :destination :sf-plus :module :sf-plus :maintenance t :time-in-ms t :output 'high :priority :min))
                     (setf (continue-check-delayed *continue*) nil))))))))))
                  


(defun update-coop-world (world) ;; only called in game mode now
  (awhen (change-drawing-mode world)
         (server-send-command (list :cmd "drawing" :value it) world)
         (setf (change-drawing-mode world) nil))
  
  ;; will be incremented in game state handler as needed
  
  (setf (update-time world) 0)
  
  (case (server-stepping world)
    (:model ;; this is synchronous
            ;; shouldn't see data on socket
            ;; unless there's a state change
            ;; or a drawing change was sent
  
     ;; Drain the socket
     (loop while (listen (usocket:socket-stream (socket world)))
         for data = (server-read-message world)
         while data
         do 
           (handle-server-message world data))
  
     (when (eq (server-mode world) 'game) ;; may have transitioned to score
       (unless (pending-step world)
         (server-send-command (list :cmd "step") world)
         (setf (pending-step world) t))))
    
    (:timer ;; this is asynchronous
            ;; we could have 0 updates (the game is lagged or we're early)
            ;; could be 1 or more if we're behind 
            ;; for now just process all of them since the last
            ;; will be the most recent.
            ;; Update time will be based on how many game
            ;; states arrived.
     
     (loop while (listen (usocket:socket-stream (socket world)))
         for data = (server-read-message world)
         while data
         do 
           (handle-server-message world data)))
     
    (t
     (print-warning "Unknown server-mode: ~S" (server-mode world))))
  (schedule-next-update world))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; The module to provide the interface, parameters, etc

(defun create-sf-module (name)
  (declare (ignore name))
  
  (define-chunks ship explosion fortress fortress-explosion bonus
    vlner pnts fortress-hit shell missile trial-end foe
    condition iff mine debris inner outer points)
  
  (define-chunks ship-explosion score  
    ready game connecting connected disconnected
    other   unknown toward away )
  
  (make-instance 'sf-plus))

(defun pre-reset-sf-module (world)
 
  ;; old
  (when (server world)
    (sfapi-disconnect (server world))
    (setf (server world) nil))
  
  (setf (game-state world) 'disconnected)
  
  (setf (responses world) nil)
  (setf (auto-response world) nil)
  
  ;; hack to make the fortress always visible
  ;; since that's the case for this task
  ;; and there are no condition screens
  
  (setf (fort-exists world) t)
  
  (setf (ship-alive world) nil)
  
  (setf (fortress-there world) nil)
  (setf (last-time world) 0)
  
  ;; track complete games
  (setf (game-count world) 0)
  (setf (game-limit world) -1)
  
  
  ;; coop (and some common items)
  
  (bt:with-recursive-lock-held ((comm-lock world))
    (when (usocket::usocket-p (socket world))
      (usocket:socket-close (socket world)))
    
    (setf (connection-state world) 'disconnected)
    (setf (data-record world) nil)
    (setf (server-mode world) nil))
  
  ;; clear out the slots as needed
  
  (bt:with-recursive-lock-held ((state-lock world))
  
    (setf (update-time world) 0)
    (setf (pending-step world) nil)
  
    ;; track deaths and kills per game
    (setf (game-stats world) (list (list 0 0)))

    ;; name for the data file
    (setf (name world) nil)

    ;; track complete games
    (setf (change-drawing-mode world) nil)))


(defun post-reset-sf-module (world)
  
  ;; old
  (setf (game-condition world) nil)
  
  ;; install a keyboard device
  (install-device '("motor" "keyboard")))


(defun delete-sf-module (world)
  ;; Kill an open connection
  
  ;; old
  (when (server world)
    (sfapi-disconnect (server world))
    (setf (server world) nil))
  
  ;; coop
  (bt:with-recursive-lock-held ((comm-lock world))
    (when (usocket::usocket-p (socket world))
      (usocket:socket-close (socket world)))))


;; respond to the game buffer queries
(defun sf-query (world buffer-name slot value)
  (cond ((eq (game-type world) :older)
         (case buffer-name
           (game (case slot
                   (state
                    (case value
                      (busy nil)
                      (free t)
                      (error nil)
                      (t (print-warning "Unknown state query ~S to game buffer" value)
                         nil)))
                   (connection-state
                      nil)
                   (game-state
                    (eq (game-state world) value))
                   (t (print-warning "Unknown query ~S ~S to the game buffer" slot value))))
         
           ))
        ((eq (game-type world) :coop)
         (bt:with-recursive-lock-held ((comm-lock world))
           (case buffer-name
             (game (case slot
                     (state
                      (case value
                        (busy nil)
                        (free t)
                        (error nil)
                        (t (print-warning "Unknown state query ~S to game buffer" value)
                           nil)))
                     (connection-state
                      (eq (connection-state world) value))
                     (game-state
                      (eq value (server-mode world)))
                     (t (print-warning "Unknown query ~S ~S to the game buffer" slot value)))))))
        (t
         (case buffer-name
             (game (case slot
                     (state
                      (case value
                        (busy nil)
                        (free t)
                        (error nil)
                        (t (print-warning "Unknown state query ~S to game buffer" value)
                           nil)))
                     (connection-state
                      nil)
                     (game-state
                      nil)
                     (t (print-warning "Unknown query ~S ~S to the game buffer" slot value))))))))
         


;; custom function for printing the buffer-status of the game buffer
(defun print-sf-connection-state ()
  (let ((world (get-module :sf-plus)))
    (cond ((eq (game-type world) :older)
           (command-output "  game-state connecting  : ~S"
                           (query-buffer 'game
                                         '(game-state connecting)))
           (command-output "  game-state connected   : ~S"
                           (query-buffer 'game
                                         '(game-state connected)))
           (command-output "  game-state disconnected: ~S"
                           (query-buffer 'game
                                         '(game-state disconnected))))
          
          ((eq (game-type world) :coop)
           
           (bt:with-recursive-lock-held ((comm-lock world))
             (with-output-to-string (s)
               (format s "  connection-state connecting  : ~S"
                 (eq (connection-state world) 'connecting))
               (format s "  connection-state connected   : ~S"
                 (eq (connection-state world) 'connected))
               (format s "  connection-state disconnected: ~S"
                 (eq (connection-state world) 'disconnected))
               (format s "  game-state ready: ~S"
                 (eq 'ready (server-mode world)))
               (format s "  game-state game: ~S"
                 (eq 'game (server-mode world)))
               (format s "  game-state score: ~S"
                 (eq 'score (server-mode world))))))
          (t
           ))))

;; function to handle the setting and getting of the module's parameters

(defun sf-params (world param)
  (bt:with-recursive-lock-held ((param-lock world))
    (cond ((consp param)
           (case (car param)
             (:sf-host (setf (sf-host world) (cdr param)))
             (:sf-port (setf (sf-port world) (cdr param)))
             (:log-sf-data-in (setf (collect-incoming world) (cdr param)))
             (:log-sf-data-out (setf (record-outgoing world) (cdr param)))
             (:sf-data-hook (setf (data-hook world) (cdr param)))
             (:coop-data-hook (setf (coop-data-hook world) (cdr param)))
             
             (:sf-server-type (setf (sf-server-type world) (cdr param)))

             (:continue-time 
              (setf (continue-time world) (seconds->ms (cdr param)))
              (cdr param))
             (:print-comm-data (setf (print-communication world) (cdr param)))
             (:sf-player (setf (server-player world) (cdr param)))
             (:game-ticks (setf (game-ticks world) (cdr param)))))
          (t
           (case param
             (:sf-host (sf-host world))
             (:sf-port (sf-port world))
             (:log-sf-data-in (collect-incoming world))
             (:log-sf-data-out (record-outgoing world))
             (:sf-data-hook (data-hook world))
             (:coop-data-hook (coop-data-hook world))
             
             (:sf-server-type (sf-server-type world))
             
             (:continue-time (ms->seconds (continue-time world)))
             (:print-comm-data (print-communication world))
             (:sf-player (server-player world))
             (:game-ticks (game-ticks world)))))))


;; define the module itself

(define-module-fct :sf-plus '((game nil nil (connection-state game-state) print-sf-connection-state))
  (list
   (define-parameter :sf-host :valid-test #'stringp :default-value "127.0.0.1"
     :warning "a string" :documentation "IP address or host name of SF server")
   (define-parameter :sf-port :valid-test 'integerp  :default-value 3000
     :warning "an integer" :documentation "Port of the server for connecting")
   (define-parameter :log-sf-data-in :valid-test #'tornil :default-value nil
     :warning "T or nil" :documentation "Record all incoming SF data")
   (define-parameter :log-sf-data-out :valid-test #'tornil :default-value nil
     :warning "T or nil" :documentation "Record all commands sent to SF")
   (define-parameter :print-comm-data :valid-test #'tornil :default-value nil
     :warning "T or nil" :documentation "Print the communication information between the model and game")
   (define-parameter :sf-data-hook :valid-test #'fctornil
     :default-value nil
     :warning "a function or nil" :documentation "Called with each data line received from the game interface for space fortress and space track")
   
   (define-parameter :coop-data-hook :valid-test #'fctornil
     :default-value nil
     :warning "a function or nil" :documentation "Called with each data line received from the game interface for coop space fortress")
   
   (define-parameter :sf-player :valid-test 'numornil :default-value nil
     :warning "a number or nil" :documentation "Enable playing Space Fortress and indicate which player")
   
   (define-parameter :continue-time :valid-test #'posnum
     :default-value 5
     :warning "a positive number" :documentation "Seconds to wait before the model automatically presses to continue between games")
   
   (define-parameter :game-ticks :valid-test 'integerp :default-value 10800
     :warning "an integer" :documentation "Number for the last tick in a game to help with synchronization")
   
   (define-parameter :sf-server-type :valid-test (lambda (x) (find x '(sfapi-tcp-client sfapi-external-process-client)))
                                     :default-value 'sfapi-tcp-client
     :warning "sfapi-tcp-client or sfapi-external-process-client" :documentation "server type to use."))

  :version "11.0a1"
  :documentation "Module to allow ACT-R 7.21 to connect and play Co-op Space Fortress and the older Space Fortress and Space Track"
  :creation 'create-sf-module
  :reset (list 'pre-reset-sf-module 'post-reset-sf-module)
  :delete 'delete-sf-module
  :query 'sf-query
  :params 'sf-params)


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Coop connection 


(defun get-model-name ()
  (if (and (boundp '*cluster-run-start*) (boundp '*cluster-run-id*))
      (progn
        (format nil "model_~d_~d" (symbol-value '*cluster-run-start*) (symbol-value '*cluster-run-id*)))
      (let (number)
        (if (probe-file (translate-logical-pathname "SF-DATA:sf-model-number.txt"))
            (with-open-file (f (translate-logical-pathname "SF-DATA:sf-model-number.txt") :direction :input)
              (setf number (ignore-errors (read f nil :done))))
            (setf number -1))
        (if (numberp number)
            (incf number)
            (setf number 0))

        (with-open-file (f (translate-logical-pathname "SF-DATA:sf-model-number.txt") :direction :output :if-exists :supersede :if-does-not-exist :create)
          (write number :stream f))
        (format nil "model_~d" number))))

(defun connect-to-coop-game (module draw stepping game bot)
  (labels ((bad-connection-exit (message)
             (print-warning message)
             (when (usocket::usocket-p (socket module))
               (usocket:socket-close (socket module))
               (setf (socket module) nil))
                                
             (setf (connection-state module) 'disconnected)
             (return-from connect-to-coop-game nil))
           
           (parse-data-line (message)
             (let ((data (server-read-message module)))
               (if data
                   data
                   (bad-connection-exit message))))
           
           (send-command (cmd error-message)
             (server-send-command cmd module)
             (let ((data (parse-data-line error-message)))
               (unless (and (equal (get-value data :msgtype) "response")
                            (equal (get-value data :cmd) (getf cmd :cmd))
                            (get-value data :success))
                 (bad-connection-exit (format nil "Did not get a successful response for command ~S: ~S" cmd data)))
               data))
           
           (set-config-parameter (key value)
             (send-command (list :cmd key :value value)
                           (format nil "Failed reading response to ~A setting" key))))
    
    (bt:with-recursive-lock-held ((comm-lock module))
      
      (if (eq (connection-state module) 'disconnected)
          
          (let ((socket (handler-case (usocket:socket-connect (sf-host module) (sf-port module) :nodelay :if-supported)
                          (error (x) (bad-connection-exit
                                      (format nil "Error ~/print-error-message/ occurred while trying to open a socket connection to host ~s port ~s" 
                                        x (sf-host module) (sf-port module)))))))
            
            (setf (socket module) socket)
            (setf (connection-state module) 'connecting)
            
            (let ((data (parse-data-line "Failed to read initial config")))
              
              (bt:with-recursive-lock-held ((state-lock module))
              
                ;; create the update times to to total a second
                (setf (game-freq module) (get-value data :fps))
                (setf (update-times module) (create-times (game-freq module)))
                
                (when (not (and (equal (get-value data :msgtype) "stateChange")
                                (equal (get-value data :state) "config")))
                  (bad-connection-exit (format nil "Expected initial config line but got ~s" data)))
                
                (set-config-parameter "drawing" (if draw 1 0))
                (set-config-parameter "player" (server-player module))
                (set-config-parameter "stepping" (ecase stepping (:model "model") (:timer "timer")))
                (when game
		  (set-config-parameter "condition" game))
		(when bot
		  (set-config-parameter "bot" 1))
		(setf (server-stepping module) stepping)
                (setf (name module) (get-model-name))
                (set-config-parameter "id" (name module))
                (send-command (list :cmd "continue") "Failed to progress from config screen")
                ;; The next message will be a state change into the
                ;; "game" state. So consider the module connected.
                (setf (connection-state module) 'connected
                  (server-mode module) 'ready)
                t)))
        (print-warning "Model ~s is already connected cannot connect." (current-model))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Functions for actually playing the game -- call these to run the model.
;;;
;;; The game can be run for a specific time in seconds using the play-sf function
;;; or a fixed number of games using the play-sf-games function.  Each of those
;;; has one required parameter for the indicated value (time or games).
;;;
;;; They also have 4 keyword parameters:
;;;
;;; :speed should be a number and indicates how fast to run the game relative
;;;        to real-time.  The default is 1 which means at real time.  A bigger
;;;        value runs the model faster than real time and a smaller value slower.
;;;
;;; :cont whether the current run should be continuted.  If this is specified
;;;       as t then it does not reconnect the model before starting running
;;;       and if it is nil then it resets the model and starts a new session.
;;;       The default is nil.
;;;
;;; :draw can be t or nil.  Indicates whether the game server displays
;;;       the game as the model plays.  A value of t means show everything and nil
;;;       means show nothing.  The default value is t.
;;;
;;; :stepping can be :timer or :model.  Determines how the model gets the game data.
;;;       In model mode the game waits for the model at each step, but in timer mode
;;;       the game just sends the updates when they happen based on the real time
;;;       clock.
;;;
;;;       When running with a person as one of the players :timer mode should be used
;;;       along with a speed of 1 (:model mode should also work with a speed of 1 for
;;;       playing with a person but it has the potential to affect the person's game).
;;;       When running two models :model mode should be used so that it can play as 
;;;       fast or slow as needed, and stepping or stopping won't affect the game time.
;;;       
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


(defun update-game-count ()
  (bt:with-lock-held ((game-counter-lock *game-counter*))
    (pushnew (current-model) (game-counter-updated-players *game-counter*))
    (when (= (length (game-counter-updated-players *game-counter*))
             (game-counter-players *game-counter*))
      (incf (game-counter-current-game *game-counter*))
      (setf (game-counter-updated-players *game-counter*) nil))))

(defun initialize-game-counter (&optional (limit 0))
  (bt:with-lock-held ((game-counter-lock *game-counter*))
    (setf (game-counter-players *game-counter*) 0)
    (setf (game-counter-updated-players *game-counter*) nil)
    (setf (game-counter-ready-players *game-counter*) nil)
    (setf (game-counter-limit *game-counter*) limit)
    (setf (game-counter-current-game *game-counter*) 0)))

(defun increment-player-count ()
  (bt:with-lock-held ((game-counter-lock *game-counter*))
    (incf (game-counter-players *game-counter*))))

(defun check-coop-game-count (x)
  ;; for safety allow it to have run over to catch for a bad update,
  ;; spurious result from the server, or other unforseen issue.
  (declare (ignore x))
  
  (bt:with-lock-held ((game-counter-lock *game-counter*))
    (>= (game-counter-current-game *game-counter*) (game-counter-limit *game-counter*))))


(defun play-sf-coop (time &key (speed 1) (cont nil) (draw t) (stepping :model)(port nil)(game nil))
  (run-sf-coop time :time speed cont draw stepping port game))

(defun play-sf-coop-games (games &key (speed 1) (cont nil) (draw t) (stepping :model)(port nil)(game nil)(bot nil))
  (run-sf-coop games :games speed cont draw stepping port game bot))


(defun run-sf-coop (value mode speed cont draw stepping port game bot)
  (if cont
      (progn
        (initialize-game-counter (if (eq mode :games) value 0))
        (dolist (m (mp-models))
          (with-model-eval m
            (let ((module (get-module :sf-plus)))
              (unless (eq (game-type module) :coop)
                (when port (sgp-fct (list :sf-port port)))
                
                (unless (connect-to-coop-game module draw stepping)
                  (print-warning "Error connecting model ~s to game." m)
                  (return-from run-sf-coop))
                
                (setf (game-type module) :coop))
              
              (when (and module (bt:with-recursive-lock-held ((param-lock module)) (server-player module)))
                (increment-player-count)
                (bt:with-recursive-lock-held ((state-lock module))
                  (setf (change-drawing-mode module) (if draw 1 0)))))))
        (if (eq mode :games)
            (run-until-condition 'check-coop-game-count speed)
          (run-full-time value speed)))
      (let (players)
        (reset)
        (initialize-game-counter (if (eq mode :games) value 0))
        (dolist (m (mp-models))
          (with-model-eval m
            (let ((module (get-module :sf-plus)))
              (setf (game-type module) :coop)
              (when port (sgp-fct (list :sf-port port)))
              (awhen (and module (bt:with-recursive-lock-held ((param-lock module)) (server-player module)))
                     (if (find it players)
                         (progn
                           (print-warning "Duplicate player number ~s for model ~s.  Cannot run Space fortress." it m)
                           (return-from run-sf-coop))
                       (progn
                         (increment-player-count)
                         (push it players)
                         (unless (connect-to-coop-game module draw stepping game bot)
                           (print-warning "Error connecting model ~s to game." m)
                           (return-from run-sf-coop))))
                     
                     (schedule-next-update module)))))
        (if (eq mode :games)
            (run-until-condition 'check-coop-game-count speed)
          (run-full-time value speed)))))



;;; Older games


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; 
;;; Handle all the config info. for the current game servers when a new session gets
;;; started with one of the running functions
;;;
;;; - read the first config line
;;; - send draw and id settings verifying that the result is t for each
;;;
;;; If anything doesn't work out punt and close the connection

(defun same-id-test (requested returned)
  (unless (stringp requested)
    (setf requested (princ-to-string requested)))
  (string-equal requested returned))

(define-condition config-error (error)
  ((msg :initform nil :initarg :msg :accessor config-error-msg))
  (:report (lambda (condition stream)
             (format stream "error during config: ~A" (config-error-msg condition)))))

(defun config-cleanup-after-error (module)
  (when (server module)
    (sfapi-disconnect (server module))
    (setf (server module) nil))
  (setf (game-state module) 'disconnected))

(defun config-parse-data-line (module message)
  (multiple-value-bind (data err)
      (ignore-errors (read-from-string (get-line-of-data-from-game module)))
    (if (subtypep (type-of err) 'condition)
        (error 'config-error :msg message)
        (progn
          (when (collect-incoming module)
            (push (list :in (mp-time) (get-internal-real-time) data) (data-record module)))
          data))))

(defun config-set-parameter (module expected-screen key value)
  (server-write-line module "~A ~A" key value)
  (let ((data (config-parse-data-line module (format nil "Failed reading response to ~A setting" key))))
    (unless (and (eq expected-screen (read-from-string (getf data :screen-type)))
                 (getf data :result))
      (error 'config-error :msg (format nil "Did not get a true config result from changing ~A -> ~A: ~S" key value data)))
    data))

(defun config-set-settings (module expected-screen &key draw condition log-format config game-number seed extra)
  (config-set-parameter module expected-screen "drawing" (if draw 1 0))
  (when condition (config-set-parameter module expected-screen "condition" condition))
  (when log-format (config-set-parameter module expected-screen "logformat" log-format))
  (when game-number (config-set-parameter module expected-screen "gnum" game-number))
  (when seed (config-set-parameter module expected-screen "seed" seed))
  (dolist (pair config)
    (config-set-parameter module expected-screen (format nil "config ~A" (first pair)) (second pair)))
  (dolist (pair extra)
    (config-set-parameter module expected-screen (format nil "extra ~A" (first pair)) (second pair))))

(defun connect-to-game (module &key draw name condition log-format config game-number seed extra)
  (when (and (eq (game-type module) :older) (not (eq (game-state module) 'disconnected)))
    (print-warning "Model is already connected cannot connect.")
    (return-from connect-to-game))
  
  (handler-case
      (progn
        (setf (server module) (make-instance (sf-server-type module)))
        
        (when (subtypep (type-of (server module)) 'sfapi-tcp-client)
          (setf (sfapi-port (server module)) (sf-port module))
          (setf (sfapi-host (server module)) (sf-host module)))

        (sfapi-connect (server module))
        
        (when (not (server module))
          (error 'config-error :msg "Could not open server connection."))
        
        (let ((data (config-parse-data-line module "Failed to read initial config")))
          (setf (server-current-screen module) (read-from-string (getf data :screen-type)))
          (unless (eq 'config (server-current-screen module))
            (error 'config-error :msg (format nil "Expected a config line but got ~s" data)))

          (config-set-settings module 'config
                               :draw draw
                               :condition condition
                               :log-format log-format
                               :config config
                               :game-number game-number
                               :seed seed
                               :extra extra)

          (let ((n (if name
                       name
                       (if (and (boundp '*cluster-run-start*) (boundp '*cluster-run-id*))
                           (progn
                             (format nil "model_~d_~d" (symbol-value '*cluster-run-start*) (symbol-value '*cluster-run-id*)))
                           (let (number)
                             (if (probe-file (translate-logical-pathname "SF-DATA:sf-model-number.txt"))
                                 (with-open-file (f (translate-logical-pathname "SF-DATA:sf-model-number.txt") :direction :input)
                                   (setf number (ignore-errors (read f nil :done))))
                                 (setf number -1))
                             (if (numberp number)
                                 (incf number)
                                 (setf number 0))
                           
                             (with-open-file (f (translate-logical-pathname "SF-DATA:sf-model-number.txt") :direction :output :if-exists :supersede :if-does-not-exist :create)
                               (write number :stream f))
                             (format nil "model_~d" number))))))
                
            (setf (name module) n)
                
            (config-set-parameter module 'config "id" n)
            
            (setf (game-type module) :older)
            (setf (game-state module) 'connected))))
    (config-error (c)
      (print-warning (config-error-msg c))
      (config-cleanup-after-error module)
      nil)))

(defun connect-to-game-in-progress (module &key draw name condition log-format config game-number seed extra)
  (when (or (not (eq (game-type module) :older)) (eq (game-state module) 'disconnected))
    (print-warning "Model must already be connected.")
    (return-from connect-to-game-in-progress))
  (handler-case
      (progn
        (when (eq (server-current-screen module) 'game)
          (server-write-line module "abort")
          (let ((data (config-parse-data-line module "Failed to abort game")))
            (setf (server-current-screen module) (read-from-string (getf data :screen-type)))))
        (unless (eq 'score (server-current-screen module))
          (error 'config-error :msg "Expected score screen.")
          (return-from connect-to-game-in-progress))
        (config-set-settings module 'score
                             :draw draw
                             :condition condition
                             :log-format log-format
                             :config config
                             :game-number game-number
                             :seed seed
                             :extra extra)
        (when name
          (setf (name module) name)
          (config-set-parameter module 'score "id" name))
        t)
    (config-error (c)
      (print-warning (config-error-msg c))
      (config-cleanup-after-error module)
      nil)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;;
;;; Functions for actually playing the game -- call these to run the model.
;;;
;;; The game can be run for a specific time in seconds using the play-sf function
;;; or a fixed number of games using the play-sf-games function.  Each of those
;;; has one required parameter for the indicated value (time or games).
;;;
;;; They also have 4 keyword parameters:
;;;
;;; :speed should be a number and indicates how fast to run the game relative
;;;        to real-time.  The default is 1 which means at real time.  A bigger 
;;;        value runs the model faster than real time and a smaller value slower.
;;;
;;; :cont whether the current run should be continuted.  If this is specified
;;;       as t then it does not reconnect the model before starting running
;;;       and if it is nil then it resets the model and starts a new session.
;;;       The default is nil.
;;;
;;; :draw can be t, nil, or :debug.  Indicates whether the game server displays
;;;       the game as the model plays.  A value of t means show everything and nil
;;;       means show nothing.  For some versions of the server :debug causes a
;;;       very limited display (sufficient to see that something is happening and
;;;       it's not hung), but not all support that mode in which case it will
;;;       show everything like t does.  The default value is t.
;;;
;;; :name should be a symbol, string, or number which is used to create a directory
;;;       in the game's data folder for the model's game data files.  If it is not
;;;       provided then the name will be "model_#" where # is one more than the 
;;;       number read from the sf-model-number.txt file found in the same directory
;;;       as the one from which this file was loaded or the number 0 if that file
;;;       is not found.  When it generates a number to use it will write that number
;;;       to the sf-model-number.txt file (creating the file if it does not exist).
;;;
;;; :condition the string which names the game condition to play.
;;;
;;; :log-format not sure...
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;;
;;; Game now runs in real-time mode to allow for speed control (doesn't use a custom
;;; clock anymore to simplify things).

(defun play-sf (time &key (speed 1) (cont nil) (draw t) name condition log-format config game-number seed port)
  
  (unless cont (reset))  
  
  (dolist (m (mp-models))
    (with-model-eval m

      (let ((module (get-module :sf-plus)))
        (if module
            (progn
              (if cont
                  (progn
                    (if (or (null condition) (equalp condition (game-condition module)))
                        (progn
                          (if (eq (game-type module) :older)
                              (setf (change-drawing-mode module) (if draw (if (eq draw :debug) 1 2) 0))
                            
                            (progn
                              (when port (sgp-fct (list :sf-port port)))
                              
                              (when (connect-to-game module :draw draw :name name :condition condition :log-format log-format :config config :game-number game-number :seed seed)
                                (setf (game-type module) :older)
                                (setf (game-condition module) condition)
                                (let* ((events (meta-p-events (current-mp)))
                                       (event (find 'update-world events :key 'act-r-event-action)))
                                  (when event
                                    (delete-event (act-r-event-num event)))
                                  (schedule-event-now 'update-world :destination :sf-plus :module :sf-plus :priority :max)))))    
                          
                          )
                      
                      (progn
                        (pre-reset-sf-module module)
                        (when port (sgp-fct (list :sf-port port)))
                        (setf (game-type module) :older)
                        (when (connect-to-game module :draw draw :name name :condition condition :log-format log-format :config config :game-number game-number :seed seed)
                          (setf (game-condition module) condition)
                          (let* ((events (meta-p-events (current-mp)))
                                 (event (find 'update-world events :key 'act-r-event-action)))
                            (when event
                              (delete-event (act-r-event-num event)))
                            (schedule-event-now 'update-world :destination :sf-plus :module :sf-plus :priority :max))
                          )))
                    )
                (progn
                  
                  (setf (game-type module) :older)
                  (when port (sgp-fct (list :sf-port port)))
                  (when (connect-to-game module :draw draw :name name :condition condition :log-format log-format :config config :game-number game-number :seed seed)
                    (setf (game-condition module) condition)
                    (schedule-event 0 'update-world :destination :sf-plus :module :sf-plus :priority :max :time-in-ms t)
                    ))))
          (print-warning "Cannot play space fortress")))))
  (run-full-time time speed))

(defun check-game-count (x) (declare (ignore x)) 
  (let ((success t))
    (dolist (m (mp-models) success)
      (with-model-eval m
        (let ((module (get-module :sf-plus)))
          (unless (>= (game-count module) (game-limit module))
            (setf success nil)))))))

(defun play-sf-games (games &key (speed 1) (cont nil) (draw t) name condition log-format config game-number seed extra port)
  
  (unless cont (reset))
  
  (dolist (m (mp-models))
    (with-model-eval m
  (let ((module (get-module :sf-plus)))
  
    (cond
      ((not module)
       (print-warning "Cannot play space fortress"))
      (cont
       
       (setf (game-count module) 0)
       (setf (game-limit module) games)
       
       (if (eq (game-type module) :older)
           (progn
             (when (connect-to-game-in-progress module :draw draw :name name :condition condition :log-format log-format :config config :game-number game-number :seed seed :extra extra)
               (let* ((events (meta-p-events (current-mp)))
                      (event (find 'update-world events :key 'act-r-event-action)))
                 (when event
                   (delete-event (act-r-event-num event)))
                 (schedule-event-now 'update-world :destination :sf-plus :module :sf-plus :priority :max))
               ))
         (progn
           (when port (sgp-fct (list :sf-port port)))
           (setf (game-type module) :older)
           (when (connect-to-game module :draw draw :name name :condition condition :log-format log-format :config config :game-number game-number :seed seed :extra extra)
             (setf (game-condition module) condition)
             (schedule-event 0 'update-world :destination :sf-plus :module :sf-plus :priority :max :time-in-ms t)
             ))))
      (t
       
       (setf (game-count module) 0)
       (setf (game-limit module) games)
       (setf (game-type module) :older)
       
       (when (connect-to-game module :draw draw :name name :condition condition :log-format log-format :config config :game-number game-number :seed seed :extra extra)
         (setf (game-condition module) condition)
         (schedule-event 0 'update-world :destination :sf-plus :module :sf-plus :priority :max :time-in-ms t)
         ))))))
  (run-until-condition 'check-game-count speed))
#|
This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
|#
